from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any

from pydantic import BaseModel, Field

from ..exceptions import ObjectValidationError, ValidationError
from ..schema import GenericSchemaAPI, RelationshipKind, RelationshipSchema
from ..yaml import InfrahubFile, InfrahubFileKind
from .models import InfrahubObjectParameters
from .processors.factory import DataProcessorFactory

if TYPE_CHECKING:
    from ..client import InfrahubClient
    from ..node import InfrahubNode
    from ..schema import MainSchemaTypesAPI, RelationshipSchema


def validate_list_of_scalars(value: list[Any]) -> bool:
    return all(isinstance(item, (str, int, float, bool)) for item in value)


def validate_list_of_hfids(value: list[Any]) -> bool:
    return all(isinstance(item, (str, list)) for item in value)


def validate_list_of_data_dicts(value: list[Any]) -> bool:
    return all(isinstance(item, dict) and "data" in item for item in value)


def validate_list_of_objects(value: list[Any]) -> bool:
    return all(isinstance(item, dict) for item in value)


class RelationshipDataFormat(str, Enum):
    UNKNOWN = "unknown"

    ONE_REF = "one_ref"
    ONE_OBJ = "one_obj"

    MANY_OBJ_DICT_LIST = "many_obj_dict_list"
    MANY_OBJ_LIST_DICT = "many_obj_list_dict"
    MANY_REF = "many_ref_list"


class RelationshipInfo(BaseModel):
    name: str
    rel_schema: RelationshipSchema
    peer_kind: str
    peer_rel: RelationshipSchema | None = None
    reason_relationship_not_valid: str | None = None
    format: RelationshipDataFormat = RelationshipDataFormat.UNKNOWN

    @property
    def is_bidirectional(self) -> bool:
        """Indicate if a relationship with the same identifier exists on the other side"""
        return bool(self.peer_rel)

    @property
    def is_mandatory(self) -> bool:
        if not self.peer_rel:
            return False
        # For hierarchical node, currently the relationship to the parent is always optional in the schema even if it's mandatory
        # In order to build the tree from top to bottom, we need to consider it as mandatory
        # While it should technically work bottom-up, it created some unexpected behavior while loading the menu
        if self.peer_rel.cardinality == "one" and self.peer_rel.kind == RelationshipKind.HIERARCHY:
            return True
        return not self.peer_rel.optional

    @property
    def is_valid(self) -> bool:
        return not self.reason_relationship_not_valid

    @property
    def is_reference(self) -> bool:
        return self.format in [RelationshipDataFormat.ONE_REF, RelationshipDataFormat.MANY_REF]

    def get_context(self, value: Any) -> dict:
        """Return a dict to insert to the context if the relationship is mandatory"""
        if self.peer_rel and self.is_mandatory and self.peer_rel.cardinality == "one":
            return {self.peer_rel.name: value}
        if self.peer_rel and self.is_mandatory and self.peer_rel.cardinality == "many":
            return {self.peer_rel.name: [value]}
        return {}

    def find_matching_relationship(
        self, peer_schema: MainSchemaTypesAPI, force: bool = False
    ) -> RelationshipSchema | None:
        """Find the matching relationship on the other side of the relationship"""
        if self.peer_rel and not force:
            return self.peer_rel

        try:
            self.peer_rel = peer_schema.get_matching_relationship(
                id=self.rel_schema.identifier or "", direction=self.rel_schema.direction
            )
        except ValueError:
            pass

        return self.peer_rel


async def get_relationship_info(
    client: InfrahubClient, schema: MainSchemaTypesAPI, name: str, value: Any, branch: str | None = None
) -> RelationshipInfo:
    """
    Get the relationship info for a given relationship name.
    """
    rel_schema = schema.get_relationship(name=name)

    info = RelationshipInfo(name=name, peer_kind=rel_schema.peer, rel_schema=rel_schema)

    if isinstance(value, dict) and "data" not in value:
        info.reason_relationship_not_valid = f"Relationship {name} must be a dict with 'data'"
        return info

    if isinstance(value, dict) and "kind" in value:
        info.peer_kind = value["kind"]

    peer_schema = await client.schema.get(kind=info.peer_kind, branch=branch)

    try:
        info.peer_rel = peer_schema.get_matching_relationship(
            id=rel_schema.identifier or "", direction=rel_schema.direction
        )
    except ValueError:
        pass

    if rel_schema.cardinality == "one" and isinstance(value, list):
        # validate the list is composed of string
        if validate_list_of_scalars(value):
            info.format = RelationshipDataFormat.ONE_REF
        else:
            info.reason_relationship_not_valid = "Too many objects provided for a relationship of cardinality one"

    elif rel_schema.cardinality == "one" and isinstance(value, str):
        info.format = RelationshipDataFormat.ONE_REF

    elif rel_schema.cardinality == "one" and isinstance(value, dict) and "data" in value:
        info.format = RelationshipDataFormat.ONE_OBJ

    elif (
        rel_schema.cardinality == "many"
        and isinstance(value, dict)
        and "data" in value
        and validate_list_of_objects(value["data"])
    ):
        # Initial format, we need to support it for backward compatibility for menu
        # it's helpful if there is only one type of object to manage
        info.format = RelationshipDataFormat.MANY_OBJ_DICT_LIST

    elif rel_schema.cardinality == "many" and isinstance(value, dict) and "data" not in value:
        info.reason_relationship_not_valid = "Invalid structure for a relationship of cardinality many,"
        " either provide a dict with data as a list or a list of objects"

    elif rel_schema.cardinality == "many" and isinstance(value, list):
        if validate_list_of_data_dicts(value):
            info.format = RelationshipDataFormat.MANY_OBJ_LIST_DICT
        elif validate_list_of_hfids(value):
            info.format = RelationshipDataFormat.MANY_REF
        else:
            info.reason_relationship_not_valid = "Invalid structure for a relationship of cardinality many,"
            " either provide a list of dict with data or a list of hfids"

    return info


class InfrahubObjectFileData(BaseModel):
    kind: str
    parameters: InfrahubObjectParameters = Field(default_factory=InfrahubObjectParameters)
    data: list[dict[str, Any]] = Field(default_factory=list)

    async def _get_processed_data(self, data: list[dict[str, Any]]) -> list[dict[str, Any]]:
        """Get data processed according to the strategy"""

        return await DataProcessorFactory.process_data(kind=self.kind, parameters=self.parameters, data=data)

    async def validate_format(self, client: InfrahubClient, branch: str | None = None) -> list[ObjectValidationError]:
        errors: list[ObjectValidationError] = []
        schema = await client.schema.get(kind=self.kind, branch=branch)

        processed_data = await self._get_processed_data(data=self.data)
        self.data = processed_data

        for idx, item in enumerate(processed_data):
            errors.extend(
                await self.validate_object(
                    client=client,
                    position=[idx + 1],
                    schema=schema,
                    data=item,
                    branch=branch,
                    default_schema_kind=self.kind,
                    parameters=self.parameters,
                )
            )
        return errors

    async def process(self, client: InfrahubClient, branch: str | None = None) -> None:
        schema = await client.schema.get(kind=self.kind, branch=branch)
        processed_data = await self._get_processed_data(data=self.data)

        for idx, item in enumerate(processed_data):
            await self.create_node(
                client=client,
                schema=schema,
                data=item,
                position=[idx + 1],
                branch=branch,
                default_schema_kind=self.kind,
                parameters=self.parameters,
            )

    @classmethod
    async def validate_object(
        cls,
        client: InfrahubClient,
        schema: MainSchemaTypesAPI,
        data: dict,
        position: list[int | str],
        context: dict | None = None,
        branch: str | None = None,
        default_schema_kind: str | None = None,
        parameters: InfrahubObjectParameters | None = None,
    ) -> list[ObjectValidationError]:
        parameters = parameters or InfrahubObjectParameters()
        errors: list[ObjectValidationError] = []
        context = context.copy() if context else {}

        # First validate if all mandatory fields are present
        for element in schema.mandatory_input_names:
            if not any([element in data.keys(), element in context.keys()]):
                errors.append(ObjectValidationError(position=position + [element], message=f"{element} is mandatory"))

        # Validate if all attributes are valid
        for key, value in data.items():
            if key not in schema.attribute_names and key not in schema.relationship_names:
                errors.append(
                    ObjectValidationError(
                        position=position + [key],
                        message=f"{key} is not a valid attribute or relationship for {schema.kind}",
                    )
                )

            if key in schema.attribute_names:
                if not isinstance(value, (str, int, float, bool, list, dict)):
                    errors.append(
                        ObjectValidationError(
                            position=position + [key],
                            message=f"{key} must be a string, int, float, bool, list, or dict",
                        )
                    )

            if key in schema.relationship_names:
                rel_info = await get_relationship_info(
                    client=client, schema=schema, name=key, value=value, branch=branch
                )
                if not rel_info.is_valid:
                    errors.append(
                        ObjectValidationError(
                            position=position + [key],
                            message=rel_info.reason_relationship_not_valid or "Invalid relationship",
                        )
                    )

                errors.extend(
                    await cls.validate_related_nodes(
                        client=client,
                        position=position + [key],
                        rel_info=rel_info,
                        data=value,
                        context=context,
                        branch=branch,
                        default_schema_kind=default_schema_kind,
                        parameters=parameters,
                    )
                )

        return errors

    @classmethod
    async def validate_related_nodes(
        cls,
        client: InfrahubClient,
        position: list[int | str],
        rel_info: RelationshipInfo,
        data: dict | list[dict],
        context: dict | None = None,
        branch: str | None = None,
        default_schema_kind: str | None = None,
        parameters: InfrahubObjectParameters | None = None,
    ) -> list[ObjectValidationError]:
        parameters = parameters or InfrahubObjectParameters()
        context = context.copy() if context else {}
        errors: list[ObjectValidationError] = []

        if isinstance(data, (list, str)) and rel_info.format == RelationshipDataFormat.ONE_REF:
            return errors

        if isinstance(data, list) and rel_info.format == RelationshipDataFormat.MANY_REF:
            return errors

        if isinstance(data, dict) and rel_info.format == RelationshipDataFormat.ONE_OBJ:
            peer_kind = data.get("kind") or rel_info.peer_kind
            peer_schema = await cls.get_peer_schema(
                client=client, peer_kind=peer_kind, branch=branch, default_schema_kind=default_schema_kind
            )

            rel_info.find_matching_relationship(peer_schema=peer_schema)
            context.update(rel_info.get_context(value="placeholder"))

            errors.extend(
                await cls.validate_object(
                    client=client,
                    position=position,
                    schema=peer_schema,
                    data=data["data"],
                    context=context,
                    branch=branch,
                    default_schema_kind=default_schema_kind,
                    parameters=parameters,
                )
            )
            return errors

        if isinstance(data, dict) and rel_info.format == RelationshipDataFormat.MANY_OBJ_DICT_LIST:
            peer_kind = data.get("kind") or rel_info.peer_kind
            peer_schema = await cls.get_peer_schema(
                client=client, peer_kind=peer_kind, branch=branch, default_schema_kind=default_schema_kind
            )

            rel_info.find_matching_relationship(peer_schema=peer_schema)
            context.update(rel_info.get_context(value="placeholder"))

            processed_data = await DataProcessorFactory.process_data(
                kind=peer_kind, data=data["data"], parameters=parameters
            )

            for idx, peer_data in enumerate(processed_data):
                context["list_index"] = idx
                errors.extend(
                    await cls.validate_object(
                        client=client,
                        position=position + [idx + 1],
                        schema=peer_schema,
                        data=peer_data,
                        context=context,
                        branch=branch,
                        default_schema_kind=default_schema_kind,
                        parameters=parameters,
                    )
                )
            return errors

        if isinstance(data, list) and rel_info.format == RelationshipDataFormat.MANY_OBJ_LIST_DICT:
            for idx, item in enumerate(data):
                context["list_index"] = idx
                peer_kind = item.get("kind") or rel_info.peer_kind
                peer_schema = await cls.get_peer_schema(
                    client=client, peer_kind=peer_kind, branch=branch, default_schema_kind=default_schema_kind
                )

                rel_info.find_matching_relationship(peer_schema=peer_schema)
                context.update(rel_info.get_context(value="placeholder"))

                errors.extend(
                    await cls.validate_object(
                        client=client,
                        position=position + [idx + 1],
                        schema=peer_schema,
                        data=item["data"],
                        context=context,
                        branch=branch,
                        default_schema_kind=default_schema_kind,
                        parameters=parameters,
                    )
                )
            return errors

        errors.append(
            ObjectValidationError(
                position=position,
                message=f"Relationship {rel_info.rel_schema.name} doesn't have the right format {rel_info.rel_schema.cardinality} / {type(data)}",
            )
        )
        return errors

    @classmethod
    def enrich_node(cls, data: dict, context: dict) -> dict:  # noqa: ARG003
        return data

    @classmethod
    async def create_node(
        cls,
        client: InfrahubClient,
        schema: MainSchemaTypesAPI,
        data: dict,
        position: list[int | str],
        context: dict | None = None,
        branch: str | None = None,
        default_schema_kind: str | None = None,
        parameters: InfrahubObjectParameters | None = None,
    ) -> InfrahubNode:
        parameters = parameters or InfrahubObjectParameters()
        context = context.copy() if context else {}

        errors = await cls.validate_object(
            client=client,
            position=position,
            schema=schema,
            data=data,
            context=context,
            branch=branch,
            default_schema_kind=default_schema_kind,
            parameters=parameters,
        )
        if errors:
            messages = [str(error) for error in errors]
            raise ObjectValidationError(position=position, message="Object is not valid - " + ", ".join(messages))

        clean_data: dict[str, Any] = {}

        # List of relationships that need to be processed after the current object has been created
        remaining_rels = []
        rels_info: dict[str, RelationshipInfo] = {}

        for key, value in data.items():
            if key in schema.attribute_names:
                clean_data[key] = value
                continue

            if key in schema.relationship_names:
                rel_info = await get_relationship_info(
                    client=client, schema=schema, name=key, value=value, branch=branch
                )
                rels_info[key] = rel_info

                if not rel_info.is_valid:
                    client.log.info(rel_info.reason_relationship_not_valid)
                    continue

                # We need to determine if the related object depend on this object or if this is the other way around.
                #  - if the relationship is bidirectional and is mandatory on the other side, then we need to create this object First
                #  - if the relationship is bidirectional and is not mandatory on the other side, then we need should create the related object First
                #  - if the relationship is not bidirectional, then we need to create the related object First
                if rel_info.is_reference and isinstance(value, list):
                    clean_data[key] = value
                elif rel_info.format == RelationshipDataFormat.ONE_REF and isinstance(value, str):
                    clean_data[key] = [value]
                elif not rel_info.is_reference and rel_info.is_bidirectional and rel_info.is_mandatory:
                    remaining_rels.append(key)
                elif not rel_info.is_reference and not rel_info.is_mandatory:
                    if rel_info.format == RelationshipDataFormat.ONE_OBJ:
                        nodes = await cls.create_related_nodes(
                            client=client,
                            position=position,
                            rel_info=rel_info,
                            data=value,
                            branch=branch,
                            default_schema_kind=default_schema_kind,
                        )
                        clean_data[key] = nodes[0]

                    else:
                        nodes = await cls.create_related_nodes(
                            client=client,
                            position=position,
                            rel_info=rel_info,
                            data=value,
                            branch=branch,
                            default_schema_kind=default_schema_kind,
                            parameters=InfrahubObjectParameters(**value.get("parameters"))
                            if "parameters" in value
                            else None,
                        )
                        clean_data[key] = nodes

                else:
                    raise ValueError(f"Situation unaccounted for: {rel_info}")

        if context:
            clean_context = {
                ckey: cvalue
                for ckey, cvalue in context.items()
                if ckey in schema.relationship_names + schema.attribute_names
            }
            clean_data.update(clean_context)

        clean_data = cls.enrich_node(data=clean_data, context=context or {})

        node = await client.create(kind=schema.kind, branch=branch, data=clean_data)
        await node.save(allow_upsert=True)

        display_label = node.get_human_friendly_id_as_string() or f"{node.get_kind()} : {node.id}"
        client.log.info(f"Created node: {display_label}")

        for rel in remaining_rels:
            context = {}

            # If there is a peer relationship, we add the node id to the context
            rel_info = rels_info[rel]
            context.update(rel_info.get_context(value=node.id))

            await cls.create_related_nodes(
                client=client,
                parent_node=node,
                rel_info=rel_info,
                position=position,
                data=data[rel],
                context=context,
                branch=branch,
                default_schema_kind=default_schema_kind,
                parameters=InfrahubObjectParameters(**data[rel].get("parameters"))
                if "parameters" in data[rel]
                else None,
            )

        return node

    @classmethod
    async def create_related_nodes(
        cls,
        client: InfrahubClient,
        rel_info: RelationshipInfo,
        position: list[int | str],
        data: dict | list[dict],
        parent_node: InfrahubNode | None = None,
        context: dict | None = None,
        branch: str | None = None,
        default_schema_kind: str | None = None,
        parameters: InfrahubObjectParameters | None = None,
    ) -> list[InfrahubNode]:
        parameters = parameters or InfrahubObjectParameters()
        nodes: list[InfrahubNode] = []
        context = context.copy() if context else {}

        if isinstance(data, dict) and rel_info.format == RelationshipDataFormat.ONE_OBJ:
            peer_kind = data.get("kind") or rel_info.peer_kind
            peer_schema = await client.schema.get(kind=peer_kind, branch=branch)

            if parent_node:
                rel_info.find_matching_relationship(peer_schema=peer_schema)
                context.update(rel_info.get_context(value=parent_node.id))

            new_node = await cls.create_node(
                client=client,
                schema=peer_schema,
                position=position,
                data=data["data"],
                context=context,
                branch=branch,
                default_schema_kind=default_schema_kind,
                parameters=parameters,
            )
            return [new_node]

        if isinstance(data, dict) and rel_info.format == RelationshipDataFormat.MANY_OBJ_DICT_LIST:
            peer_kind = data.get("kind") or rel_info.peer_kind
            peer_schema = await cls.get_peer_schema(
                client=client, peer_kind=peer_kind, branch=branch, default_schema_kind=default_schema_kind
            )

            if parent_node:
                rel_info.find_matching_relationship(peer_schema=peer_schema)
                context.update(rel_info.get_context(value=parent_node.id))

            expanded_data = await DataProcessorFactory.process_data(
                kind=peer_kind, data=data["data"], parameters=parameters
            )

            for idx, peer_data in enumerate(expanded_data):
                context["list_index"] = idx
                if isinstance(peer_data, dict):
                    node = await cls.create_node(
                        client=client,
                        schema=peer_schema,
                        position=position + [rel_info.name, idx + 1],
                        data=peer_data,
                        context=context,
                        branch=branch,
                        default_schema_kind=default_schema_kind,
                        parameters=parameters,
                    )
                    nodes.append(node)
            return nodes

        if isinstance(data, list) and rel_info.format == RelationshipDataFormat.MANY_OBJ_LIST_DICT:
            for idx, item in enumerate(data):
                context["list_index"] = idx

                peer_kind = item.get("kind") or rel_info.peer_kind
                peer_schema = await cls.get_peer_schema(
                    client=client, peer_kind=peer_kind, branch=branch, default_schema_kind=default_schema_kind
                )

                if parent_node:
                    rel_info.find_matching_relationship(peer_schema=peer_schema)
                    context.update(rel_info.get_context(value=parent_node.id))

                node = await cls.create_node(
                    client=client,
                    schema=peer_schema,
                    position=position + [rel_info.name, idx + 1],
                    data=item["data"],
                    context=context,
                    branch=branch,
                    default_schema_kind=default_schema_kind,
                    parameters=parameters,
                )
                nodes.append(node)

            return nodes

        raise ValueError(
            f"Relationship {rel_info.rel_schema.name} doesn't have the right format {rel_info.rel_schema.cardinality} / {type(data)}"
        )

    @classmethod
    async def get_peer_schema(
        cls, client: InfrahubClient, peer_kind: str, branch: str | None = None, default_schema_kind: str | None = None
    ) -> MainSchemaTypesAPI:
        peer_schema = await client.schema.get(kind=peer_kind, branch=branch)
        if not isinstance(peer_schema, GenericSchemaAPI):
            return peer_schema

        if not default_schema_kind:
            raise ValueError(f"Found a peer schema as a generic {peer_kind} but no default value was provided")

        # if the initial peer_kind was a generic, we try the default_schema_kind
        peer_schema = await client.schema.get(kind=default_schema_kind, branch=branch)
        if isinstance(peer_schema, GenericSchemaAPI):
            raise ValueError(f"Default schema kind {default_schema_kind} can't be a generic")
        return peer_schema


class ObjectFile(InfrahubFile):
    _spec: InfrahubObjectFileData | None = None

    @property
    def spec(self) -> InfrahubObjectFileData:
        if not self._spec:
            try:
                self._spec = InfrahubObjectFileData(**self.data.spec)
            except Exception as exc:
                raise ValidationError(identifier=str(self.location), message=str(exc))
        return self._spec

    def validate_content(self) -> None:
        super().validate_content()
        if self.kind != InfrahubFileKind.OBJECT:
            raise ValueError("File is not an Infrahub Object file")
        try:
            self._spec = InfrahubObjectFileData(**self.data.spec)
        except Exception as exc:
            raise ValidationError(identifier=str(self.location), message=str(exc))

    async def validate_format(self, client: InfrahubClient, branch: str | None = None) -> None:
        self.validate_content()
        errors = await self.spec.validate_format(client=client, branch=branch)
        if errors:
            raise ValidationError(identifier=str(self.location), messages=[str(error) for error in errors])

    async def process(self, client: InfrahubClient, branch: str | None = None) -> None:
        await self.spec.process(client=client, branch=branch)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import Boolean, InputObjectType, Mutation, String
from graphql import GraphQLResolveInfo
from opentelemetry import trace
from typing_extensions import Self

from infrahub.core.manager import NodeManager
from infrahub.core.schema import ProfileSchema
from infrahub.graphql.types.context import ContextInput
from infrahub.log import get_logger
from infrahub.profiles.node_applier import NodeProfilesApplier
from infrahub.workflows.catalogue import PROFILE_REFRESH_MULTIPLE

from .main import InfrahubMutationMixin, InfrahubMutationOptions

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.database import InfrahubDatabase
    from infrahub.graphql.initialization import GraphqlContext
    from infrahub.services.adapters.workflow import InfrahubWorkflow

log = get_logger()


class InfrahubProfileMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(
        cls,
        schema: ProfileSchema,
        _meta: InfrahubMutationOptions | None = None,
        **options: dict[str, Any],
    ) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, ProfileSchema):
            raise ValueError(f"You need to pass a valid ProfileSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)
        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    async def _send_profile_refresh_workflows(
        cls,
        db: InfrahubDatabase,
        workflow_service: InfrahubWorkflow,
        branch_name: str,
        obj: Node,
        node_ids: list[str] | None = None,
    ) -> None:
        if not node_ids:
            related_nodes = await obj.related_nodes.get_relationships(db=db)  # type: ignore[attr-defined]
            node_ids = [rel.peer_id for rel in related_nodes]
        if node_ids:
            await workflow_service.submit_workflow(
                workflow=PROFILE_REFRESH_MULTIPLE,
                parameters={
                    "branch_name": branch_name,
                    "node_ids": node_ids,
                },
            )

    @classmethod
    def _get_profile_attr_values_map(cls, obj: Node) -> dict[str, Any]:
        attr_values_map = {}
        for attr_schema in obj.get_schema().attributes:
            # profile name update can be ignored
            if attr_schema.name == "profile_name":
                continue
            attr_values_map[attr_schema.name] = getattr(obj, attr_schema.name).value
        return attr_values_map

    @classmethod
    async def _get_profile_related_node_ids(cls, db: InfrahubDatabase, obj: Node) -> set[str]:
        related_nodes = await obj.related_nodes.get_relationships(db=db)  # type: ignore[attr-defined]
        if related_nodes:
            related_node_ids = {rel.peer_id for rel in related_nodes}
        else:
            related_node_ids = set()
        return related_node_ids

    @classmethod
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,
        override_data: dict[str, Any] | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context
        db = database or graphql_context.db
        workflow_service = graphql_context.active_service.workflow

        obj, mutation = await super().mutate_create(
            info=info, data=data, branch=branch, database=database, override_data=override_data
        )
        await cls._send_profile_refresh_workflows(
            db=db, workflow_service=workflow_service, branch_name=branch.name, obj=obj
        )

        return obj, mutation

    @classmethod
    async def _call_mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        db: InfrahubDatabase,
        obj: Node,
        skip_uniqueness_check: bool = False,
    ) -> tuple[Node, Self]:
        workflow_service = info.context.active_service.workflow
        original_attr_values = cls._get_profile_attr_values_map(obj=obj)
        original_related_node_ids = await cls._get_profile_related_node_ids(db=db, obj=obj)

        obj, mutation = await super()._call_mutate_update(
            info=info, data=data, branch=branch, db=db, obj=obj, skip_uniqueness_check=skip_uniqueness_check
        )

        updated_attr_values = cls._get_profile_attr_values_map(obj=obj)
        updated_related_node_ids = await cls._get_profile_related_node_ids(db=db, obj=obj)

        if original_attr_values != updated_attr_values:
            await cls._send_profile_refresh_workflows(
                db=db, workflow_service=workflow_service, branch_name=branch.name, obj=obj
            )
        elif updated_related_node_ids != original_related_node_ids:
            removed_node_ids = original_related_node_ids - updated_related_node_ids
            added_node_ids = updated_related_node_ids - original_related_node_ids
            await cls._send_profile_refresh_workflows(
                db=db,
                workflow_service=workflow_service,
                branch_name=branch.name,
                obj=obj,
                node_ids=list(removed_node_ids) + list(added_node_ids),
            )

        return obj, mutation

    @classmethod
    async def _delete_obj(cls, graphql_context: GraphqlContext, branch: Branch, obj: Node) -> list[Node]:
        db = graphql_context.db
        workflow_service = graphql_context.active_service.workflow
        related_node_ids = await cls._get_profile_related_node_ids(db=db, obj=obj)
        deleted = await super()._delete_obj(graphql_context=graphql_context, branch=branch, obj=obj)
        await cls._send_profile_refresh_workflows(
            db=db, workflow_service=workflow_service, branch_name=branch.name, obj=obj, node_ids=list(related_node_ids)
        )
        return deleted


class ProfilesRefreshInput(InputObjectType):
    id = String(required=False)


class InfrahubProfilesRefresh(Mutation):
    class Arguments:
        data = ProfilesRefreshInput(required=True)
        context = ContextInput(required=False)

    ok = Boolean()

    @classmethod
    @trace.get_tracer(__name__).start_as_current_span("profiles_refresh")
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: ProfilesRefreshInput,
        context: ContextInput | None = None,  # noqa: ARG003
    ) -> Self:
        graphql_context: GraphqlContext = info.context
        db = graphql_context.db
        branch = graphql_context.branch
        obj = await NodeManager.get_one(
            db=db,
            branch=branch,
            id=str(data.id),
            include_source=True,
        )
        node_profiles_applier = NodeProfilesApplier(db=db, branch=branch)
        updated_fields = await node_profiles_applier.apply_profiles(node=obj)
        if updated_fields:
            await obj.save(db=db, fields=updated_fields)

        return cls(ok=True)



from __future__ import annotations

from collections import defaultdict
from itertools import chain
from typing import TYPE_CHECKING, Any

import ujson
from rich.progress import Progress, TaskID

from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.constants import GLOBAL_BRANCH_NAME, BranchSupportType, RelationshipDirection
from infrahub.core.initialization import get_root_node
from infrahub.core.migrations.shared import MigrationResult, get_migration_console
from infrahub.core.query import Query, QueryType
from infrahub.core.schema import NodeSchema
from infrahub.exceptions import SchemaNotFoundError
from infrahub.types import is_large_attribute_type

from ..shared import MigrationRequiringRebase
from .load_schema_branch import get_or_load_schema_branch

if TYPE_CHECKING:
    from infrahub.core.schema import AttributeSchema, NodeSchema, ProfileSchema, TemplateSchema
    from infrahub.core.schema.basenode_schema import SchemaAttributePath
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.database import InfrahubDatabase


console = get_migration_console()


class DefaultBranchNodeCount(Query):
    """
    Get the number of Node vertices on the given branches that are not in the kinds_to_skip list
    Only works for default and global branches. Non-default branches would only return a count of nodes
    created on the given branches
    """

    name = "get_branch_node_count"
    type = QueryType.READ

    def __init__(self, kinds_to_skip: list[str], **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.kinds_to_skip = kinds_to_skip

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params = {
            "branch_names": [registry.default_branch, GLOBAL_BRANCH_NAME],
            "kinds_to_skip": self.kinds_to_skip,
        }
        query = """
MATCH (n:Node)-[e:IS_PART_OF]->(:Root)
WHERE NOT n.kind IN $kinds_to_skip
AND e.branch IN $branch_names
AND e.status = "active"
AND e.to IS NULL
AND NOT exists((n)-[:IS_PART_OF {branch: e.branch, status: "deleted"}]->(:Root))
WITH count(*) AS num_nodes
        """
        self.add_to_query(query)
        self.return_labels = ["num_nodes"]

    def get_num_nodes(self) -> int:
        result = self.get_result()
        if not result:
            return 0
        return result.get_as_type(label="num_nodes", return_type=int)


class GetResultMapQuery(Query):
    def get_result_map(self, schema_paths: list[SchemaAttributePath]) -> dict[str, list[str | None]]:
        """
        Get the values for the given schema paths for all the Nodes captured by this query
        """
        # the query results for attribute and schema paths are unordered
        # so we make this list of keys for ordering the results from the query
        schema_path_keys: list[tuple[str, RelationshipDirection, str] | str] = []
        for schema_path in schema_paths:
            if schema_path.is_type_attribute and schema_path.attribute_schema:
                path_key: str | tuple[str, RelationshipDirection, str] = schema_path.attribute_schema.name
            elif schema_path.is_type_relationship and schema_path.relationship_schema and schema_path.attribute_schema:
                path_key = (
                    schema_path.relationship_schema.get_identifier(),
                    schema_path.relationship_schema.direction,
                    schema_path.attribute_schema.name,
                )
            schema_path_keys.append(path_key)

        result_map: dict[str, list[str | None]] = {}
        for result in self.get_results():
            node_uuid = result.get_as_type(label="n_uuid", return_type=str)

            # for each node, build a map of the schema path key to value so that they
            # can be ordered correctly for the input `schema_paths`
            schema_path_value_map: dict[str | tuple[str, RelationshipDirection, str], Any] = {}
            attr_values_tuples: list[tuple[str, Any]] = result.get_as_type(label="attr_vals_list", return_type=list)
            for attr_value_tuple in attr_values_tuples:
                attr_name = attr_value_tuple[0]
                attr_value = attr_value_tuple[1]
                schema_path_value_map[attr_name] = attr_value

            relationship_values_tuples: list[tuple[str, str, str, Any]] = result.get_as_type(
                label="peer_attr_vals_list", return_type=list
            )
            for rel_value_tuple in relationship_values_tuples:
                rel_name = rel_value_tuple[0]
                direction_raw = rel_value_tuple[1]
                direction = RelationshipDirection.BIDIR
                match direction_raw:
                    case "outbound":
                        direction = RelationshipDirection.OUTBOUND
                    case "inbound":
                        direction = RelationshipDirection.INBOUND
                peer_attr_name = rel_value_tuple[2]
                peer_val = rel_value_tuple[3]
                schema_path_value_map[rel_name, direction, peer_attr_name] = peer_val

            schema_path_values: list[str | None] = []
            for schema_path_key in schema_path_keys:
                value = schema_path_value_map.get(schema_path_key)
                schema_path_values.append(str(value) if value is not None else None)
            result_map[node_uuid] = schema_path_values
        return result_map


class GetPathDetailsBranchQuery(GetResultMapQuery):
    name = "get_path_details_branch"
    type = QueryType.READ
    insert_limit = False

    def __init__(
        self, schema_kind: str, schema_paths: list[SchemaAttributePath], updates_only: bool = True, **kwargs: Any
    ) -> None:
        super().__init__(**kwargs)

        if self.branch.name in [registry.default_branch, GLOBAL_BRANCH_NAME]:
            raise ValueError("This query can only be used on non-default branches")
        self.schema_kind = schema_kind
        self.updates_only = updates_only
        self.attribute_names = []
        self.bidir_rel_attr_map: dict[str, list[str]] = defaultdict(list)
        self.outbound_rel_attr_map: dict[str, list[str]] = defaultdict(list)
        self.inbound_rel_attr_map: dict[str, list[str]] = defaultdict(list)
        for schema_path in schema_paths:
            if schema_path.is_type_attribute and schema_path.attribute_schema:
                self.attribute_names.append(schema_path.attribute_schema.name)
            elif schema_path.is_type_relationship and schema_path.relationship_schema and schema_path.attribute_schema:
                key = schema_path.relationship_schema.get_identifier()
                value = schema_path.attribute_schema.name
                if schema_path.relationship_schema.direction is RelationshipDirection.BIDIR:
                    self.bidir_rel_attr_map[key].append(value)
                elif schema_path.relationship_schema.direction is RelationshipDirection.OUTBOUND:
                    self.outbound_rel_attr_map[key].append(value)
                elif schema_path.relationship_schema.direction is RelationshipDirection.INBOUND:
                    self.inbound_rel_attr_map[key].append(value)

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        branch_filter, branch_filter_params = self.branch.get_query_filter_path(at=self.at)
        self.params.update(branch_filter_params)
        self.params.update(
            {
                "branch_name": self.branch.name,
                "attribute_names": self.attribute_names,
                "outbound_rel_ids": list(self.outbound_rel_attr_map.keys()),
                "inbound_rel_ids": list(self.inbound_rel_attr_map.keys()),
                "bidirectional_rel_ids": list(self.bidir_rel_attr_map.keys()),
                "outbound_rel_attr_map": self.outbound_rel_attr_map,
                "inbound_rel_attr_map": self.inbound_rel_attr_map,
                "bidirectional_rel_attr_map": self.bidir_rel_attr_map,
                "offset": self.offset,
                "limit": self.limit,
            }
        )
        get_active_nodes_query = """
// ------------
// Get the active nodes of the given kind on the branches
// ------------
MATCH (n:%(schema_kind)s)-[r:IS_PART_OF]->(:Root)
WHERE %(branch_filter)s
WITH DISTINCT n
CALL (n) {
    MATCH (n)-[r:IS_PART_OF]->(:Root)
    WHERE %(branch_filter)s
    RETURN r.status = "active" AS is_active
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
}
WITH n, is_active
WHERE is_active = TRUE
        """ % {"schema_kind": self.schema_kind, "branch_filter": branch_filter}
        self.add_to_query(get_active_nodes_query)

        if self.updates_only:
            updated_nodes_filter_query = """
// ------------
// filter to any nodes that might have changes on the branch we care about
// ------------
OPTIONAL MATCH (n)-[r1:HAS_ATTRIBUTE]->(attr:Attribute)-[r2:HAS_VALUE]->(attr_val:AttributeValue)
WHERE attr.name in $attribute_names
AND r2.branch = $branch_name
AND r2.status = "active"
AND r2.to IS NULL
WITH n, attr_val IS NOT NULL AS has_attr_update
OPTIONAL MATCH (n)-[r1:IS_RELATED]-(rel:Relationship)-[r2:IS_RELATED]-(peer:Node)-[r3:HAS_ATTRIBUTE]-(attr:Attribute)-[r4:HAS_VALUE]->(attr_val)
WHERE rel.name IN $bidirectional_rel_ids + $outbound_rel_ids + $inbound_rel_ids
AND (
    attr.name IN $outbound_rel_attr_map[rel.name]
    OR attr.name IN $inbound_rel_attr_map[rel.name]
    OR attr.name IN $bidirectional_rel_attr_map[rel.name]
)
AND $branch_name IN [r1.branch, r2.branch, r3.branch, r4.branch]
WITH n, has_attr_update, attr_val IS NOT NULL AS has_rel_update
WITH n, any(x IN collect(has_attr_update OR has_rel_update) WHERE x = TRUE) AS has_update
WITH n, has_update
WHERE has_update = TRUE
            """
            self.add_to_query(updated_nodes_filter_query)

        get_node_details_query = """
// ------------
// Order and limit the Nodes
// ------------
ORDER BY elementId(n)
SKIP toInteger($offset)
LIMIT toInteger($limit)
// ------------
// for every possibly updated node
// get all the attribute values on this branch
// ------------
OPTIONAL MATCH (n)-[r:HAS_ATTRIBUTE]->(attr:Attribute)
WHERE attr.name IN $attribute_names
WITH DISTINCT n, attr
CALL (n, attr) {
    OPTIONAL MATCH (n)-[r1:HAS_ATTRIBUTE]->(attr)-[r2:HAS_VALUE]->(attr_val)
    WHERE all(r in [r1, r2] WHERE %(branch_filter)s)
    RETURN attr_val.value AS attr_value, r1.status = "active" AND r2.status = "active" AS is_active
    ORDER BY r2.branch_level DESC, r2.from DESC, r2.status ASC, r1.branch_level DESC, r1.from DESC, r1.status ASC
    LIMIT 1
}
WITH n, attr, attr_value
WHERE is_active = TRUE
WITH n, collect([attr.name, attr_value]) AS attr_vals_list
// ------------
// for every possibly updated node
// get all the relationships on this branch
// ------------
OPTIONAL MATCH (n)-[:IS_RELATED]-(rel:Relationship)
WHERE rel.name IN $bidirectional_rel_ids + $outbound_rel_ids + $inbound_rel_ids
WITH DISTINCT n, attr_vals_list, rel
CALL (n, rel) {
    OPTIONAL MATCH (n)-[r1:IS_RELATED]-(rel)-[r2:IS_RELATED]-(peer:Node)
    WHERE all(r in [r1, r2] WHERE %(branch_filter)s)
    AND (
        (startNode(r1) = n AND startNode(r2) = rel AND rel.name IN $outbound_rel_ids)
        OR (startNode(r1) = rel AND startNode(r2) = peer AND rel.name IN $inbound_rel_ids)
        OR (startNode(r1) = n AND startNode(r2) = peer AND rel.name IN $bidirectional_rel_ids)
    )
    RETURN
        peer,
        r1.status = "active" AND r2.status = "active" AS is_active,
        CASE
            WHEN startNode(r1) = n AND startNode(r2) = rel AND rel.name IN $outbound_rel_ids THEN "outbound"
            WHEN startNode(r1) = rel AND startNode(r2) = peer AND rel.name IN $inbound_rel_ids THEN "inbound"
            ELSE "bidir"
        END AS direction
    ORDER BY r2.branch_level DESC, r2.from DESC, r2.status ASC, r1.branch_level DESC, r1.from DESC, r1.status ASC
    LIMIT 1
}
// ------------
// get the attribute values that we care about for each relationship
// ------------
WITH n, attr_vals_list, rel.name AS rel_name, direction, peer
WHERE is_active = TRUE OR rel_name IS NULL
WITH *, CASE
    WHEN direction = "outbound" THEN $outbound_rel_attr_map[rel_name]
    WHEN direction = "inbound" THEN $inbound_rel_attr_map[rel_name]
    ELSE $bidirectional_rel_attr_map[rel_name]
END AS peer_attr_names
UNWIND COALESCE(peer_attr_names, [NULL]) AS peer_attr_name
CALL (rel_name, direction, peer, peer_attr_name){
    OPTIONAL MATCH (peer)-[r1:HAS_ATTRIBUTE]->(attr:Attribute)-[r2:HAS_VALUE]->(attr_val)
    WHERE attr.name = peer_attr_name
    AND all(r in [r1, r2] WHERE %(branch_filter)s)
    RETURN attr_val.value AS peer_attr_value, r1.status = "active" AND r2.status = "active" AS is_active
    ORDER BY r2.branch_level DESC, r2.from DESC, r2.status ASC, r1.branch_level DESC, r1.from DESC, r1.status ASC
    LIMIT 1
}
// ------------
// collect everything to return a pair of lists with each node UUID
// ------------
WITH DISTINCT n, attr_vals_list, rel_name, peer, direction, peer_attr_name, peer_attr_value
WITH n, attr_vals_list, collect([rel_name, direction, peer_attr_name, peer_attr_value]) AS peer_attr_vals_list
        """ % {"branch_filter": branch_filter}
        self.add_to_query(get_node_details_query)
        self.return_labels = ["n.uuid AS n_uuid", "attr_vals_list", "peer_attr_vals_list"]


class GetPathDetailsDefaultBranch(GetResultMapQuery):
    """
    Get the values of the given schema paths for the given kind of node on the default and global branches
    Supports limit and offset for pagination
    """

    name = "get_path_details_default_branch"
    type = QueryType.READ
    insert_limit = False

    def __init__(self, schema_kind: str, schema_paths: list[SchemaAttributePath], **kwargs: Any) -> None:
        super().__init__(**kwargs)

        self.branch_names = [registry.default_branch, GLOBAL_BRANCH_NAME]
        self.schema_kind = schema_kind
        self.attribute_names = []
        self.bidir_rel_attr_map: dict[str, list[str]] = defaultdict(list)
        self.outbound_rel_attr_map: dict[str, list[str]] = defaultdict(list)
        self.inbound_rel_attr_map: dict[str, list[str]] = defaultdict(list)
        for schema_path in schema_paths:
            if schema_path.is_type_attribute and schema_path.attribute_schema:
                self.attribute_names.append(schema_path.attribute_schema.name)
            elif schema_path.is_type_relationship and schema_path.relationship_schema and schema_path.attribute_schema:
                key = schema_path.relationship_schema.get_identifier()
                value = schema_path.attribute_schema.name
                if schema_path.relationship_schema.direction is RelationshipDirection.BIDIR:
                    self.bidir_rel_attr_map[key].append(value)
                elif schema_path.relationship_schema.direction is RelationshipDirection.OUTBOUND:
                    self.outbound_rel_attr_map[key].append(value)
                elif schema_path.relationship_schema.direction is RelationshipDirection.INBOUND:
                    self.inbound_rel_attr_map[key].append(value)

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params = {
            "branch_names": self.branch_names,
            "attribute_names": self.attribute_names,
            "outbound_rel_ids": list(self.outbound_rel_attr_map.keys()),
            "inbound_rel_ids": list(self.inbound_rel_attr_map.keys()),
            "bidirectional_rel_ids": list(self.bidir_rel_attr_map.keys()),
            "outbound_rel_attr_map": self.outbound_rel_attr_map,
            "inbound_rel_attr_map": self.inbound_rel_attr_map,
            "bidirectional_rel_attr_map": self.bidir_rel_attr_map,
            "offset": self.offset,
            "limit": self.limit,
        }
        get_details_query = """
// ------------
// Get the active nodes of the given kind on the branches
// ------------
MATCH (n:%(schema_kind)s)-[e:IS_PART_OF]->(:Root)
WHERE e.branch IN $branch_names
AND e.to IS NULL
AND e.status = "active"
// ------------
// Order and limit the Nodes
// ------------
WITH DISTINCT n
ORDER BY elementId(n)
SKIP toInteger($offset)
LIMIT toInteger($limit)
// ------------
// Get the values for the attribute schema paths of the Nodes, if any
// ------------
OPTIONAL MATCH (n)-[e:HAS_ATTRIBUTE]->(attr:Attribute)
WHERE attr.name IN $attribute_names
AND e.branch IN $branch_names
AND e.to IS NULL
AND e.status = "active"
WITH n, attr
OPTIONAL MATCH (attr)-[e:HAS_VALUE]->(attr_val:AttributeValue)
WHERE e.branch IN $branch_names
AND e.to IS NULL
AND e.status = "active"
WITH n, collect([attr.name, attr_val.value]) AS attr_vals_list
// ------------
// Get the values for the relationship schema paths of the Nodes, if any
// ------------
OPTIONAL MATCH (n)-[e1:IS_RELATED]-(rel:Relationship)-[e2:IS_RELATED]-(peer:Node)
WHERE rel.name IN $bidirectional_rel_ids + $outbound_rel_ids + $inbound_rel_ids
AND e1.branch IN $branch_names
AND e1.to IS NULL
AND e1.status = "active"
AND e2.branch IN $branch_names
AND e2.to IS NULL
AND e2.status = "active"
AND (
    (startNode(e1) = n AND startNode(e2) = rel AND rel.name IN $outbound_rel_ids)
    OR (startNode(e1) = rel AND startNode(e2) = peer AND rel.name IN $inbound_rel_ids)
    OR (startNode(e1) = n AND startNode(e2) = peer AND rel.name IN $bidirectional_rel_ids)
)
WITH DISTINCT n, attr_vals_list, rel.name AS rel_name, peer,  CASE
    WHEN startNode(e1) = n AND startNode(e2) = rel AND rel.name IN $outbound_rel_ids THEN "outbound"
    WHEN startNode(e1) = rel AND startNode(e2) = peer AND rel.name IN $inbound_rel_ids THEN "inbound"
    ELSE "bidir"
END AS direction
OPTIONAL MATCH (peer)-[e1:HAS_ATTRIBUTE]->(attr:Attribute)-[e2:HAS_VALUE]->(peer_attr_val:AttributeValue)
WHERE (
    (direction = "outbound" AND attr.name IN $outbound_rel_attr_map[rel_name])
    OR (direction = "inbound" AND attr.name IN $inbound_rel_attr_map[rel_name])
    OR (direction = "bidir" AND attr.name IN $bidirectional_rel_attr_map[rel_name])
)
AND e1.branch IN $branch_names
AND e1.to IS NULL
AND e1.status = "active"
AND e2.branch IN $branch_names
AND e2.to IS NULL
AND e2.status = "active"
// ------------
// collect everything to return a pair of lists with each node UUID
// ------------
WITH DISTINCT n, attr_vals_list, rel_name, peer, direction, attr.name AS peer_attr_name, peer_attr_val.value AS peer_val
WITH n, attr_vals_list, collect([rel_name, direction, peer_attr_name, peer_val]) AS peer_attr_vals_list
        """ % {"schema_kind": self.schema_kind}
        self.add_to_query(get_details_query)
        self.return_labels = ["n.uuid AS n_uuid", "attr_vals_list", "peer_attr_vals_list"]


class UpdateAttributeValuesQuery(Query):
    """
    Update the values of the given attribute schema for the input Node-id-to-value map
    Includes special handling for updating large-type attributes b/c they are not indexed and will be slow to update
    on large data sets
    """

    name = "update_attribute_values"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, attribute_schema: AttributeSchema, values_by_id_map: dict[str, Any], **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.attribute_name = attribute_schema.name
        self.is_large_type_attribute = is_large_attribute_type(attribute_schema.kind)
        self.is_branch_agnostic = attribute_schema.get_branch() is BranchSupportType.AGNOSTIC
        self.values_by_id_map = values_by_id_map

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params = {
            "node_uuids": list(self.values_by_id_map.keys()),
            "attribute_name": self.attribute_name,
            "values_by_id": self.values_by_id_map,
            "default_branch": registry.default_branch,
            "global_branch": GLOBAL_BRANCH_NAME,
            "branch": GLOBAL_BRANCH_NAME if self.is_branch_agnostic else self.branch.name,
            "branch_level": 1 if self.is_branch_agnostic else self.branch.hierarchy_level,
            "at": self.at.to_string(),
        }
        branch_filter, branch_filter_params = self.branch.get_query_filter_path(at=self.at)
        self.params.update(branch_filter_params)

        if self.is_large_type_attribute:
            # we make our own index of value to database ID, creating any vertices that are missing
            # the mapping is a list of tuples instead of an actual mapping b/c creating an actual map is not possible
            # without apoc functions
            all_distinct_values = list(set(self.values_by_id_map.values()))
            diy_index_query = """
MATCH (av:AttributeValue&!AttributeValueIndexed {is_default: false})
WHERE av.value IN $all_distinct_values
WITH collect([av.value, elementId(av)]) AS value_id_pairs, collect(av.value) AS found_values
WITH value_id_pairs, found_values,
    reduce(
        missing_distinct_values = [], value IN $all_distinct_values |
            CASE
                WHEN value IN found_values THEN missing_distinct_values
                ELSE missing_distinct_values + [value]
            END
    ) AS missing_distinct_values
CALL (missing_distinct_values) {
    UNWIND missing_distinct_values AS missing_value
    CREATE (av:AttributeValue {is_default: false, value: missing_value})
    RETURN collect([av.value, elementId(av)]) AS created_value_id_pairs
}
WITH value_id_pairs + created_value_id_pairs AS value_id_pairs
            """
            self.params["all_distinct_values"] = all_distinct_values
        else:
            # if this is not a large-type attribute, then just set the map to be empty
            diy_index_query = """WITH [] AS value_id_pairs"""

        self.add_to_query(diy_index_query)

        if self.branch.name in [registry.default_branch, GLOBAL_BRANCH_NAME]:
            update_value_query = """
// ------------
// Find the Nodes and Attributes we need to update
// ------------
MATCH (n:Node)-[e:IS_PART_OF]->(:Root)
WHERE n.uuid IN $node_uuids
AND e.branch IN [$default_branch, $global_branch]
AND e.to IS NULL
AND e.status = "active"
WITH DISTINCT n, value_id_pairs
MATCH (n)-[e:HAS_ATTRIBUTE]->(attr:Attribute {name: $attribute_name})
WHERE e.branch IN [$default_branch, $global_branch]
AND e.to IS NULL
AND e.status = "active"
// ------------
// If the attribute has an existing value on the branch, then set the to time on it
// but only if the value is different from the new value
// ------------
WITH DISTINCT n, attr, value_id_pairs
CALL (attr) {
    OPTIONAL MATCH (attr)-[e:HAS_VALUE]->(existing_av)
    WHERE e.branch IN [$default_branch, $global_branch]
    AND e.to IS NULL
    AND e.status = "active"
    RETURN existing_av, e AS existing_has_value
}
CALL (existing_has_value) {
    WITH existing_has_value
    WHERE existing_has_value IS NOT NULL
    SET existing_has_value.to = $at
}
            """
        else:
            update_value_query = """
// ------------
// Find the Nodes and Attributes we need to update
// ------------
MATCH (n:Node)
WHERE n.uuid IN $node_uuids
CALL (n) {
    MATCH (n)-[r:IS_PART_OF]->(:Root)
    WHERE %(branch_filter)s
    RETURN r.status = "active" AS is_active
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
}
WITH n, value_id_pairs, is_active
WHERE is_active = TRUE
WITH DISTINCT n, value_id_pairs
CALL (n) {
    MATCH (n)-[r:HAS_ATTRIBUTE]->(attr:Attribute {name: $attribute_name})
    WHERE %(branch_filter)s
    RETURN attr, r.status = "active"  AS is_active
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
}
WITH DISTINCT n, attr, value_id_pairs, is_active
WHERE is_active = TRUE
// ------------
// If the attribute has an existing value on the branch, then set the to time on it
// but only if the value is different from the new value
// ------------
CALL (n, attr) {
    OPTIONAL MATCH (attr)-[r:HAS_VALUE]->(existing_av)
    WHERE %(branch_filter)s
    WITH r, existing_av
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
    WITH CASE
        WHEN existing_av.value <> $values_by_id[n.uuid]
        AND r.status = "active"
        AND r.branch = $branch
        THEN [r, existing_av]
        ELSE [NULL, NULL]
    END AS existing_details
    RETURN existing_details[0] AS existing_has_value, existing_details[1] AS existing_av
}
CALL (existing_has_value) {
    WITH existing_has_value
    WHERE existing_has_value IS NOT NULL
    SET existing_has_value.to = $at
}
            """ % {"branch_filter": branch_filter}
        self.add_to_query(update_value_query)

        if self.is_large_type_attribute:
            # use the index we created at the start to get the database ID of the AttributeValue vertex
            # and then link the Attribute to the AttributeValue
            set_value_query = """
// ------------
// only make updates if the existing value is not the same as the new value
// ------------
WITH attr, existing_av, value_id_pairs, $values_by_id[n.uuid] AS required_value
WHERE existing_av.value <> required_value
OR existing_av IS NULL
WITH attr, value_id_pairs, required_value,
    reduce(av_vertex_id = NULL, pair IN value_id_pairs |
        CASE
            WHEN av_vertex_id IS NOT NULL THEN av_vertex_id
            WHEN pair[0] = required_value THEN pair[1]
            ELSE av_vertex_id
        END
    ) AS av_vertex_id
MATCH (av:AttributeValue)
WHERE elementId(av) = av_vertex_id
CREATE (attr)-[r:HAS_VALUE { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(av)
            """
        else:
            # if not a large-type attribute, then we can just use the regular MERGE clause
            # that makes use of the index on AttributeValueIndexed
            set_value_query = """
// ------------
// only make updates if the existing value is not the same as the new value
// ------------
WITH n, attr, existing_av, value_id_pairs, $values_by_id[n.uuid] AS required_value
WHERE existing_av.value <> required_value
OR existing_av IS NULL
CALL (n, attr) {
    MERGE (av:AttributeValue&AttributeValueIndexed {is_default: false, value: $values_by_id[n.uuid]} )
    WITH av, attr
    LIMIT 1
    CREATE (attr)-[r:HAS_VALUE { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(av)
}
            """
        self.add_to_query(set_value_query)


class Migration044(MigrationRequiringRebase):
    """
    Backfill `human_friendly_id` and `display_label` attributes for nodes with schemas that define them.
    """

    name: str = "044_backfill_hfid_display_label_in_db"
    minimum_version: int = 43
    update_batch_size: int = 1000
    # skip these b/c the attributes on these schema-related nodes are used to define the values included in
    # the human_friendly_id and display_label attributes on instances of these schema, so should not be updated
    kinds_to_skip: list[str] = ["SchemaNode", "SchemaAttribute", "SchemaRelationship", "SchemaGeneric"]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def _do_one_schema_all(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        schema: NodeSchema | ProfileSchema | TemplateSchema,
        schema_branch: SchemaBranch,
        attribute_schema_map: dict[AttributeSchema, AttributeSchema],
        progress: Progress | None = None,
        update_task: TaskID | None = None,
    ) -> None:
        print(f"Processing {schema.kind}...", end="")

        schema_paths_by_name: dict[str, list[SchemaAttributePath]] = {}
        for source_attribute_schema in attribute_schema_map.keys():
            node_schema_property = getattr(schema, source_attribute_schema.name)
            if not node_schema_property:
                continue
            if isinstance(node_schema_property, list):
                schema_paths_by_name[source_attribute_schema.name] = [
                    schema.parse_schema_path(path=str(path), schema=schema_branch) for path in node_schema_property
                ]
            else:
                schema_paths_by_name[source_attribute_schema.name] = [
                    schema.parse_schema_path(path=str(node_schema_property), schema=schema_branch)
                ]
        all_schema_paths = list(chain(*schema_paths_by_name.values()))
        offset = 0

        # loop until we get no results from the get_details_query
        while True:
            if branch.is_default:
                get_details_query: GetResultMapQuery = await GetPathDetailsDefaultBranch.init(
                    db=db,
                    schema_kind=schema.kind,
                    schema_paths=all_schema_paths,
                    offset=offset,
                    limit=self.update_batch_size,
                )
            else:
                get_details_query = await GetPathDetailsBranchQuery.init(
                    db=db,
                    branch=branch,
                    schema_kind=schema.kind,
                    schema_paths=all_schema_paths,
                    updates_only=False,
                    offset=offset,
                    limit=self.update_batch_size,
                )
            await get_details_query.execute(db=db)

            num_updates = 0
            for source_attribute_schema, destination_attribute_schema in attribute_schema_map.items():
                schema_paths = schema_paths_by_name[source_attribute_schema.name]
                schema_path_values_map = get_details_query.get_result_map(schema_paths)
                num_updates = max(num_updates, len(schema_path_values_map))
                formatted_schema_path_values_map = {}
                for k, v in schema_path_values_map.items():
                    if not v:
                        continue
                    if destination_attribute_schema.kind == "List":
                        formatted_schema_path_values_map[k] = ujson.dumps(v)
                    else:
                        formatted_schema_path_values_map[k] = " ".join(item for item in v if item is not None)

                if not formatted_schema_path_values_map:
                    continue

                update_display_label_query = await UpdateAttributeValuesQuery.init(
                    db=db,
                    branch=branch,
                    attribute_schema=destination_attribute_schema,
                    values_by_id_map=formatted_schema_path_values_map,
                )
                await update_display_label_query.execute(db=db)

            if progress is not None and update_task is not None:
                progress.update(update_task, advance=num_updates)

            if num_updates == 0:
                break

            offset += self.update_batch_size

        print("done")

    async def execute(self, db: InfrahubDatabase) -> MigrationResult:
        root_node = await get_root_node(db=db, initialize=False)
        default_branch_name = root_node.default_branch
        default_branch = await Branch.get_by_name(db=db, name=default_branch_name)

        main_schema_branch = await get_or_load_schema_branch(db=db, branch=default_branch)

        total_nodes_query = await DefaultBranchNodeCount.init(db=db, kinds_to_skip=self.kinds_to_skip)
        await total_nodes_query.execute(db=db)
        total_nodes_count = total_nodes_query.get_num_nodes()

        base_node_schema = main_schema_branch.get("SchemaNode", duplicate=False)
        display_label_attribute_schema = base_node_schema.get_attribute("display_label")
        display_labels_attribute_schema = base_node_schema.get_attribute("display_labels")
        hfid_attribute_schema = base_node_schema.get_attribute("human_friendly_id")

        try:
            with Progress(console=console) as progress:
                update_task = progress.add_task(
                    f"Set display_label and human_friendly_id for {total_nodes_count} nodes on default branch",
                    total=total_nodes_count,
                )
                for node_schema_name in main_schema_branch.node_names:
                    if node_schema_name in self.kinds_to_skip:
                        continue

                    node_schema = main_schema_branch.get_node(name=node_schema_name, duplicate=False)
                    attribute_schema_map = {}
                    if node_schema.display_labels:
                        attribute_schema_map[display_labels_attribute_schema] = display_label_attribute_schema
                    if node_schema.human_friendly_id:
                        attribute_schema_map[hfid_attribute_schema] = hfid_attribute_schema
                    if not attribute_schema_map:
                        continue

                    await self._do_one_schema_all(
                        db=db,
                        branch=default_branch,
                        schema=node_schema,
                        schema_branch=main_schema_branch,
                        attribute_schema_map=attribute_schema_map,
                        progress=progress,
                        update_task=update_task,
                    )

        except Exception as exc:
            return MigrationResult(errors=[str(exc)])
        return MigrationResult()

    async def _do_one_schema_branch(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        schema: NodeSchema | ProfileSchema | TemplateSchema,
        schema_branch: SchemaBranch,
        source_attribute_schema: AttributeSchema,
        destination_attribute_schema: AttributeSchema,
    ) -> None:
        print(f"Processing {schema.kind}.{destination_attribute_schema.name} for {branch.name}...", end="")

        schema_property = getattr(schema, source_attribute_schema.name)
        if isinstance(schema_property, list):
            schema_paths = [
                schema.parse_schema_path(path=str(path_part), schema=schema_branch) for path_part in schema_property
            ]
        else:
            schema_paths = [schema.parse_schema_path(path=str(schema_property), schema=schema_branch)]

        offset = 0

        while True:
            # loop until we get no results from the get_details_query
            get_details_query = await GetPathDetailsBranchQuery.init(
                db=db,
                branch=branch,
                schema_kind=schema.kind,
                schema_paths=schema_paths,
                offset=offset,
                limit=self.update_batch_size,
            )
            await get_details_query.execute(db=db)

            schema_path_values_map = get_details_query.get_result_map(schema_paths)
            if not schema_path_values_map:
                print("done")
                break
            formatted_schema_path_values_map = {}
            for k, v in schema_path_values_map.items():
                if not v:
                    continue
                if destination_attribute_schema.kind == "List":
                    formatted_v = ujson.dumps(v)
                else:
                    formatted_v = " ".join(item for item in v if item is not None)
                formatted_schema_path_values_map[k] = formatted_v

            update_attr_values_query = await UpdateAttributeValuesQuery.init(
                db=db,
                branch=branch,
                attribute_schema=destination_attribute_schema,
                values_by_id_map=formatted_schema_path_values_map,
            )
            await update_attr_values_query.execute(db=db)

            offset += self.update_batch_size

    async def execute_against_branch(self, db: InfrahubDatabase, branch: Branch) -> MigrationResult:
        default_branch = await Branch.get_by_name(db=db, name=registry.default_branch)
        main_schema_branch = await get_or_load_schema_branch(db=db, branch=default_branch)
        schema_branch = await get_or_load_schema_branch(db=db, branch=branch)

        base_node_schema = schema_branch.get("SchemaNode", duplicate=False)
        display_label_attribute_schema = base_node_schema.get_attribute("display_label")
        display_labels_attribute_schema = base_node_schema.get_attribute("display_labels")
        hfid_attribute_schema = base_node_schema.get_attribute("human_friendly_id")

        try:
            for node_schema_name in schema_branch.node_names:
                if node_schema_name in self.kinds_to_skip:
                    continue

                node_schema = schema_branch.get_node(name=node_schema_name, duplicate=False)
                try:
                    default_node_schema = main_schema_branch.get_node(name=node_schema_name, duplicate=False)
                except SchemaNotFoundError:
                    default_node_schema = None
                schemas_for_universal_update_map = {}
                schemas_for_targeted_update_map = {}
                if node_schema.display_label:
                    if default_node_schema is None or default_node_schema.display_label != node_schema.display_label:
                        schemas_for_universal_update_map[display_labels_attribute_schema] = (
                            display_label_attribute_schema
                        )
                    else:
                        schemas_for_targeted_update_map[display_labels_attribute_schema] = (
                            display_label_attribute_schema
                        )

                if node_schema.human_friendly_id:
                    if (
                        default_node_schema is None
                        or default_node_schema.human_friendly_id != node_schema.human_friendly_id
                    ):
                        schemas_for_universal_update_map[hfid_attribute_schema] = hfid_attribute_schema
                    else:
                        schemas_for_targeted_update_map[hfid_attribute_schema] = hfid_attribute_schema

                if schemas_for_universal_update_map:
                    await self._do_one_schema_all(
                        db=db,
                        branch=branch,
                        schema=node_schema,
                        schema_branch=schema_branch,
                        attribute_schema_map=schemas_for_universal_update_map,
                    )

                if not schemas_for_targeted_update_map:
                    continue

                for source_attribute_schema, destination_attribute_schema in schemas_for_targeted_update_map.items():
                    await self._do_one_schema_branch(
                        db=db,
                        branch=branch,
                        schema=node_schema,
                        schema_branch=schema_branch,
                        source_attribute_schema=source_attribute_schema,
                        destination_attribute_schema=destination_attribute_schema,
                    )

        except Exception as exc:
            return MigrationResult(errors=[str(exc)])
        return MigrationResult()



import uuid

from prometheus_client import Gauge

from infrahub import __version__ as infrahub_version

WORKER_IDENTITY = str(uuid.uuid4())

INFO_METRIC = Gauge(
    "infrahub_info",
    "Information about this Infrahub instance",
    labelnames=["version", "worker_id"],
)
INFO_METRIC.labels(version=infrahub_version, worker_id=WORKER_IDENTITY).set(1)



from infrahub.core.validators.attribute.kind import AttributeKindChecker
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class SchemaAttributeKindConstraintDependency(DependencyBuilder[AttributeKindChecker]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> AttributeKindChecker:
        return AttributeKindChecker(db=context.db, branch=context.branch)



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub import lock
from infrahub.core import registry
from infrahub.core.constants import GLOBAL_BRANCH_NAME
from infrahub.graphql.registry import registry as graphql_registry
from infrahub.log import get_logger
from infrahub.worker import WORKER_IDENTITY

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.database import InfrahubDatabase

log = get_logger()


def update_graphql_schema(branch: Branch, schema_branch: SchemaBranch) -> None:
    """
    Update the GraphQL schema for the given branch.
    """

    gqlm = graphql_registry.get_manager_for_branch(branch=branch, schema_branch=schema_branch)
    gqlm.get_graphql_schema(
        include_query=True,
        include_mutation=True,
        include_subscription=True,
        include_types=True,
    )


async def create_branch_registry(db: InfrahubDatabase, branch: Branch) -> None:
    """Create a new entry in the registry for a given branch."""

    log.info("New branch detected, pulling schema", branch=branch.name, worker=WORKER_IDENTITY)
    await registry.schema.load_schema(db=db, branch=branch)
    registry.branch[branch.name] = branch
    schema_branch = registry.schema.get_schema_branch(name=branch.name)
    update_graphql_schema(branch=branch, schema_branch=schema_branch)


async def update_branch_registry(db: InfrahubDatabase, branch: Branch) -> None:
    """Update the registry for a branch if the schema hash has changed or the branch was rebased."""

    existing_branch: Branch = registry.branch[branch.name]

    if not existing_branch.schema_hash:
        log.warning("Branch schema hash is not set, cannot update branch registry")
        return

    if existing_branch.schema_hash.main == branch.active_schema_hash.main:
        log.debug(
            "Branch schema hash is the same, no need to refresh the GraphQL schema within the registry",
            branch=branch.name,
            hash=existing_branch.schema_hash.main,
            worker=WORKER_IDENTITY,
        )
        if existing_branch.branched_from != branch.branched_from:
            # If the hash is the same but the branched_from timestamp differs it means
            # that the branch has been rebased and these timestamps need to be refreshed
            # in the registry even though the schema doesn't need to be reloaded.
            log.info(
                "Updating branched_from property in registry for rebased branch",
                branch=branch.name,
                worker=WORKER_IDENTITY,
            )
            registry.branch[branch.name] = branch
        elif existing_branch.status != branch.status:
            log.info(f"Updating registry branch cache for {branch.name=}")
            registry.branch[branch.name] = branch
        return

    log.info(
        "New hash detected",
        branch=branch.name,
        hash_current=existing_branch.schema_hash.main,
        hash_new=branch.active_schema_hash.main,
        worker=WORKER_IDENTITY,
    )
    await registry.schema.load_schema(db=db, branch=branch)
    registry.branch[branch.name] = branch
    schema_branch = registry.schema.get_schema_branch(name=branch.name)

    update_graphql_schema(branch=branch, schema_branch=schema_branch)


async def refresh_branches(db: InfrahubDatabase) -> None:
    """Pull all the branches from the database and update the registry.

    If a branch is already present with a different value for the hash
    We pull the new schema from the database and we update the registry.
    """

    async with lock.registry.local_schema_lock():
        active_branches = await registry.branch_object.get_list(db=db)
        for active_branch in active_branches:
            if active_branch.name == GLOBAL_BRANCH_NAME:
                # Avoid processing updates for the global branch as it doesn't
                # have an associated schema
                continue

            if active_branch.name in registry.branch:
                await update_branch_registry(db=db, branch=active_branch)
            else:
                await create_branch_registry(db=db, branch=active_branch)

        purged_branches = await registry.purge_inactive_branches(db=db, active_branches=active_branches)
        purged_branches.update(
            graphql_registry.purge_inactive(active_branches=[branch.name for branch in active_branches])
        )
        for branch_name in sorted(purged_branches):
            log.info(f"Removed branch {branch_name!r} from the registry", branch=branch_name, worker=WORKER_IDENTITY)



from __future__ import annotations

from collections import defaultdict
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

from ..exceptions import (
    Error,
    UninitializedError,
)
from ..types import Order
from .constants import PROPERTIES_FLAG, PROPERTIES_OBJECT
from .related_node import RelatedNode, RelatedNodeSync

if TYPE_CHECKING:
    from ..client import InfrahubClient, InfrahubClientSync
    from ..schema import RelationshipSchemaAPI
    from .node import InfrahubNode, InfrahubNodeSync


class RelationshipManagerBase:
    """Base class for RelationshipManager and RelationshipManagerSync"""

    def __init__(self, name: str, branch: str, schema: RelationshipSchemaAPI) -> None:
        """
        Args:
            name (str): The name of the relationship.
            branch (str): The branch where the relationship resides.
            schema (RelationshipSchema): The schema of the relationship.
        """
        self.initialized: bool = False
        self._has_update: bool = False
        self.name = name
        self.schema = schema
        self.branch = branch

        self._properties_flag = PROPERTIES_FLAG
        self._properties_object = PROPERTIES_OBJECT
        self._properties = self._properties_flag + self._properties_object

        self.peers: list[RelatedNode | RelatedNodeSync] = []

    @property
    def peer_ids(self) -> list[str]:
        return [peer.id for peer in self.peers if peer.id]

    @property
    def peer_hfids(self) -> list[list[Any]]:
        return [peer.hfid for peer in self.peers if peer.hfid]

    @property
    def peer_hfids_str(self) -> list[str]:
        return [peer.hfid_str for peer in self.peers if peer.hfid_str]

    @property
    def has_update(self) -> bool:
        return self._has_update

    def _generate_input_data(self, allocate_from_pool: bool = False) -> list[dict]:
        return [peer._generate_input_data(allocate_from_pool=allocate_from_pool) for peer in self.peers]

    def _generate_mutation_query(self) -> dict[str, Any]:
        # Does nothing for now
        return {}

    @classmethod
    def _generate_query_data(cls, peer_data: dict[str, Any] | None = None, property: bool = False) -> dict:
        """Generates the basic structure of a GraphQL query for relationships with multiple nodes.

        Args:
            peer_data (dict[str, Union[Any, Dict]], optional): Additional data to be included in the query for each node.
                This is used to add extra fields when prefetching related node data in many-to-many relationships.

        Returns:
            Dict: A dictionary representing the basic structure of a GraphQL query for multiple related nodes.
                It includes count, edges, and node information (ID, display label, and typename), along with additional properties
                and any peer_data provided.
        """
        data: dict[str, Any] = {
            "count": None,
            "edges": {"node": {"id": None, "hfid": None, "display_label": None, "__typename": None}},
        }

        properties: dict[str, Any] = {}
        if property:
            for prop_name in PROPERTIES_FLAG:
                properties[prop_name] = None
            for prop_name in PROPERTIES_OBJECT:
                properties[prop_name] = {"id": None, "display_label": None, "__typename": None}

        if properties:
            data["edges"]["properties"] = properties
        if peer_data:
            data["edges"]["node"].update(peer_data)

        return data


class RelationshipManager(RelationshipManagerBase):
    """Manages relationships of a node in an asynchronous context."""

    def __init__(
        self,
        name: str,
        client: InfrahubClient,
        node: InfrahubNode,
        branch: str,
        schema: RelationshipSchemaAPI,
        data: Any | dict,
    ) -> None:
        """
        Args:
            name (str): The name of the relationship.
            client (InfrahubClient): The client used to interact with the backend.
            node (InfrahubNode): The node to which the relationship belongs.
            branch (str): The branch where the relationship resides.
            schema (RelationshipSchema): The schema of the relationship.
            data (Union[Any, dict]): Initial data for the relationships.
        """
        self.client = client
        self.node = node

        super().__init__(name=name, schema=schema, branch=branch)

        self.initialized = data is not None
        self._has_update = False

        if data is None:
            return

        if isinstance(data, list):
            for item in data:
                self.peers.append(
                    RelatedNode(name=name, client=self.client, branch=self.branch, schema=schema, data=item)
                )
        elif isinstance(data, dict) and "edges" in data:
            for item in data["edges"]:
                self.peers.append(
                    RelatedNode(name=name, client=self.client, branch=self.branch, schema=schema, data=item)
                )
        else:
            raise ValueError(f"Unexpected format for {name} found a {type(data)}, {data}")

    def __getitem__(self, item: int) -> RelatedNode:
        return self.peers[item]  # type: ignore[return-value]

    async def fetch(self) -> None:
        if not self.initialized:
            exclude = self.node._schema.relationship_names + self.node._schema.attribute_names
            exclude.remove(self.schema.name)
            node = await self.client.get(
                kind=self.node._schema.kind,
                id=self.node.id,
                branch=self.branch,
                include=[self.schema.name],
                exclude=exclude,
            )
            rm = getattr(node, self.schema.name)
            self.peers = rm.peers
            self.initialized = True

        ids_per_kind_map = defaultdict(list)
        for peer in self.peers:
            if not peer.id or not peer.typename:
                raise Error("Unable to fetch the peer, id and/or typename are not defined")
            ids_per_kind_map[peer.typename].append(peer.id)

        batch = await self.client.create_batch()
        for kind, ids in ids_per_kind_map.items():
            batch.add(
                task=self.client.filters,
                kind=kind,
                ids=ids,
                populate_store=True,
                branch=self.branch,
                parallel=True,
                order=Order(disable=True),
            )

        async for _ in batch.execute():
            pass

    def add(self, data: str | RelatedNode | dict) -> None:
        """Add a new peer to this relationship."""
        if not self.initialized:
            raise UninitializedError("Must call fetch() on RelationshipManager before editing members")
        new_node = RelatedNode(schema=self.schema, client=self.client, branch=self.branch, data=data)

        if (new_node.id and new_node.id not in self.peer_ids) or (
            new_node.hfid and new_node.hfid not in self.peer_hfids
        ):
            self.peers.append(new_node)
            self._has_update = True

    def extend(self, data: Iterable[str | RelatedNode | dict]) -> None:
        """Add new peers to this relationship."""
        for d in data:
            self.add(d)

    def remove(self, data: str | RelatedNode | dict) -> None:
        if not self.initialized:
            raise UninitializedError("Must call fetch() on RelationshipManager before editing members")
        node_to_remove = RelatedNode(schema=self.schema, client=self.client, branch=self.branch, data=data)

        if node_to_remove.id and node_to_remove.id in self.peer_ids:
            idx = self.peer_ids.index(node_to_remove.id)
            if self.peers[idx].id != node_to_remove.id:
                raise IndexError(f"Unexpected situation, the node with the index {idx} should be {node_to_remove.id}")

            self.peers.pop(idx)
            self._has_update = True

        elif node_to_remove.hfid and node_to_remove.hfid in self.peer_hfids:
            idx = self.peer_hfids.index(node_to_remove.hfid)
            if self.peers[idx].hfid != node_to_remove.hfid:
                raise IndexError(f"Unexpected situation, the node with the index {idx} should be {node_to_remove.hfid}")

            self.peers.pop(idx)
            self._has_update = True


class RelationshipManagerSync(RelationshipManagerBase):
    """Manages relationships of a node in a synchronous context."""

    def __init__(
        self,
        name: str,
        client: InfrahubClientSync,
        node: InfrahubNodeSync,
        branch: str,
        schema: RelationshipSchemaAPI,
        data: Any | dict,
    ) -> None:
        """
        Args:
            name (str): The name of the relationship.
            client (InfrahubClientSync): The client used to interact with the backend synchronously.
            node (InfrahubNodeSync): The node to which the relationship belongs.
            branch (str): The branch where the relationship resides.
            schema (RelationshipSchema): The schema of the relationship.
            data (Union[Any, dict]): Initial data for the relationships.
        """
        self.client = client
        self.node = node

        super().__init__(name=name, schema=schema, branch=branch)

        self.initialized = data is not None
        self._has_update = False

        if data is None:
            return

        if isinstance(data, list):
            for item in data:
                self.peers.append(
                    RelatedNodeSync(name=name, client=self.client, branch=self.branch, schema=schema, data=item)
                )
        elif isinstance(data, dict) and "edges" in data:
            for item in data["edges"]:
                self.peers.append(
                    RelatedNodeSync(name=name, client=self.client, branch=self.branch, schema=schema, data=item)
                )
        else:
            raise ValueError(f"Unexpected format for {name} found a {type(data)}, {data}")

    def __getitem__(self, item: int) -> RelatedNodeSync:
        return self.peers[item]  # type: ignore[return-value]

    def fetch(self) -> None:
        if not self.initialized:
            exclude = self.node._schema.relationship_names + self.node._schema.attribute_names
            exclude.remove(self.schema.name)
            node = self.client.get(
                kind=self.node._schema.kind,
                id=self.node.id,
                branch=self.branch,
                include=[self.schema.name],
                exclude=exclude,
            )
            rm = getattr(node, self.schema.name)
            self.peers = rm.peers
            self.initialized = True

        ids_per_kind_map = defaultdict(list)
        for peer in self.peers:
            if not peer.id or not peer.typename:
                raise Error("Unable to fetch the peer, id and/or typename are not defined")
            ids_per_kind_map[peer.typename].append(peer.id)

        batch = self.client.create_batch()
        for kind, ids in ids_per_kind_map.items():
            batch.add(
                task=self.client.filters,
                kind=kind,
                ids=ids,
                populate_store=True,
                branch=self.branch,
                parallel=True,
                order=Order(disable=True),
            )

        for _ in batch.execute():
            pass

    def add(self, data: str | RelatedNodeSync | dict) -> None:
        """Add a new peer to this relationship."""
        if not self.initialized:
            raise UninitializedError("Must call fetch() on RelationshipManager before editing members")
        new_node = RelatedNodeSync(schema=self.schema, client=self.client, branch=self.branch, data=data)

        if (new_node.id and new_node.id not in self.peer_ids) or (
            new_node.hfid and new_node.hfid not in self.peer_hfids
        ):
            self.peers.append(new_node)
            self._has_update = True

    def extend(self, data: Iterable[str | RelatedNodeSync | dict]) -> None:
        """Add new peers to this relationship."""
        for d in data:
            self.add(d)

    def remove(self, data: str | RelatedNodeSync | dict) -> None:
        if not self.initialized:
            raise UninitializedError("Must call fetch() on RelationshipManager before editing members")
        node_to_remove = RelatedNodeSync(schema=self.schema, client=self.client, branch=self.branch, data=data)

        if node_to_remove.id and node_to_remove.id in self.peer_ids:
            idx = self.peer_ids.index(node_to_remove.id)
            if self.peers[idx].id != node_to_remove.id:
                raise IndexError(f"Unexpected situation, the node with the index {idx} should be {node_to_remove.id}")
            self.peers.pop(idx)
            self._has_update = True

        elif node_to_remove.hfid and node_to_remove.hfid in self.peer_hfids:
            idx = self.peer_hfids.index(node_to_remove.hfid)
            if self.peers[idx].hfid != node_to_remove.hfid:
                raise IndexError(f"Unexpected situation, the node with the index {idx} should be {node_to_remove.hfid}")

            self.peers.pop(idx)
            self._has_update = True



from typing import ClassVar

from pydantic import Field

from .constants import EVENT_NAMESPACE
from .models import InfrahubEvent


class ArtifactEvent(InfrahubEvent):
    """Event generated when an artifact has been created or updated."""

    node_id: str = Field(..., description="The ID of the artifact")
    artifact_definition_id: str = Field(..., description="The ID of the artifact definition")
    artifact_definition_name: str = Field(..., description="The name of the artifact definition")
    target_id: str = Field(..., description="The ID of the target of the artifact")
    target_kind: str = Field(..., description="The kind of the target of the artifact")
    checksum: str = Field(..., description="The current checksum of the artifact")
    checksum_previous: str | None = Field(default=None, description="The previous checksum of the artifact")
    storage_id: str = Field(..., description="The current storage id of the artifact")
    storage_id_previous: str | None = Field(default=None, description="The previous storage id of the artifact")

    def get_related(self) -> list[dict[str, str]]:
        related = super().get_related()
        related.append(
            {
                "prefect.resource.id": self.target_id,
                "prefect.resource.role": "infrahub.related.node",
                "infrahub.node.kind": self.target_kind,
            }
        )
        related.append(
            {
                "prefect.resource.id": self.target_id,
                "prefect.resource.role": "infrahub.artifact",
                "infrahub.artifact.checksum": self.checksum,
                "infrahub.artifact.checksum_previous": self.checksum_previous or "",
                "infrahub.artifact.storage_id": self.storage_id,
                "infrahub.artifact.storage_id_previous": self.storage_id_previous or "",
                "infrahub.artifact.artifact_definition_id": self.artifact_definition_id,
                "infrahub.artifact.artifact_definition_name": self.artifact_definition_name,
            }
        )

        return related

    def get_resource(self) -> dict[str, str]:
        return {
            "prefect.resource.id": self.node_id,
            "infrahub.node.kind": "CoreArtifact",
            "infrahub.node.id": self.node_id,
            "infrahub.branch.name": self.meta.context.branch.name,
        }


class ArtifactCreatedEvent(ArtifactEvent):
    """Event generated when an artifact has been created"""

    event_name: ClassVar[str] = f"{EVENT_NAMESPACE}.artifact.created"
    infrahub_node_kind_event: ClassVar[bool] = True


class ArtifactUpdatedEvent(ArtifactEvent):
    """Event generated when an artifact has been updated"""

    event_name: ClassVar[str] = f"{EVENT_NAMESPACE}.artifact.updated"
    infrahub_node_kind_event: ClassVar[bool] = True



from __future__ import annotations

import importlib
import inspect
import logging
from pathlib import Path
from typing import TYPE_CHECKING

import typer
from infrahub_sdk.async_typer import AsyncTyper
from rich import print as rprint

from infrahub import config
from infrahub.patch.edge_adder import PatchPlanEdgeAdder
from infrahub.patch.edge_deleter import PatchPlanEdgeDeleter
from infrahub.patch.edge_updater import PatchPlanEdgeUpdater
from infrahub.patch.plan_reader import PatchPlanReader
from infrahub.patch.plan_writer import PatchPlanWriter
from infrahub.patch.queries.base import PatchQuery
from infrahub.patch.runner import (
    PatchPlanEdgeDbIdTranslator,
    PatchRunner,
)
from infrahub.patch.vertex_adder import PatchPlanVertexAdder
from infrahub.patch.vertex_deleter import PatchPlanVertexDeleter
from infrahub.patch.vertex_updater import PatchPlanVertexUpdater

from .constants import ERROR_BADGE, SUCCESS_BADGE

if TYPE_CHECKING:
    from infrahub.cli.context import CliContext
    from infrahub.database import InfrahubDatabase


patch_app = AsyncTyper(help="Commands for planning, applying, and reverting database patches")


def get_patch_runner(db: InfrahubDatabase) -> PatchRunner:
    return PatchRunner(
        plan_writer=PatchPlanWriter(),
        plan_reader=PatchPlanReader(),
        edge_db_id_translator=PatchPlanEdgeDbIdTranslator(),
        vertex_adder=PatchPlanVertexAdder(db=db),
        vertex_deleter=PatchPlanVertexDeleter(db=db),
        vertex_updater=PatchPlanVertexUpdater(db=db),
        edge_adder=PatchPlanEdgeAdder(db=db),
        edge_deleter=PatchPlanEdgeDeleter(db=db),
        edge_updater=PatchPlanEdgeUpdater(db=db),
    )


@patch_app.command(name="plan")
async def plan_patch_cmd(
    ctx: typer.Context,
    patch_path: str = typer.Argument(
        help="Path to the file containing the PatchQuery instance to run. Use Python-style dot paths, such as infrahub.cli.patch.queries.base"
    ),
    patch_plans_dir: Path = typer.Option(Path("infrahub-patches"), help="Path to patch plans directory"),  # noqa: B008
    apply: bool = typer.Option(False, help="Apply the patch immediately after creating it"),
    config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG"),
) -> None:
    """Create a plan for a given patch and save it in the patch plans directory to be applied/reverted"""
    logging.getLogger("infrahub").setLevel(logging.WARNING)
    logging.getLogger("neo4j").setLevel(logging.ERROR)
    logging.getLogger("prefect").setLevel(logging.ERROR)

    patch_module = importlib.import_module(patch_path)
    patch_query_class = None
    patch_query_class_count = 0
    for _, cls in inspect.getmembers(patch_module, inspect.isclass):
        if issubclass(cls, PatchQuery) and cls is not PatchQuery:
            patch_query_class = cls
            patch_query_class_count += 1

    patch_query_path = f"{PatchQuery.__module__}.{PatchQuery.__name__}"
    if patch_query_class is None:
        rprint(f"{ERROR_BADGE} No subclass of {patch_query_path} found in {patch_path}")
        raise typer.Exit(1)
    if patch_query_class_count > 1:
        rprint(
            f"{ERROR_BADGE} Multiple subclasses of {patch_query_path} found in {patch_path}. Please only define one per file."
        )
        raise typer.Exit(1)

    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)

    patch_query_instance = patch_query_class(db=dbdriver)
    async with dbdriver.start_session() as db:
        patch_runner = get_patch_runner(db=db)
        patch_plan_dir = await patch_runner.prepare_plan(patch_query_instance, directory=Path(patch_plans_dir))
        rprint(f"{SUCCESS_BADGE} Patch plan created at {patch_plan_dir}")
        if apply:
            await patch_runner.apply(patch_plan_directory=patch_plan_dir)
            rprint(f"{SUCCESS_BADGE} Patch plan successfully applied")

    await dbdriver.close()


@patch_app.command(name="apply")
async def apply_patch_cmd(
    ctx: typer.Context,
    patch_plan_dir: Path = typer.Argument(help="Path to the directory containing a patch plan"),
    config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG"),
) -> None:
    """Apply a given patch plan"""
    logging.getLogger("infrahub").setLevel(logging.WARNING)
    logging.getLogger("neo4j").setLevel(logging.ERROR)
    logging.getLogger("prefect").setLevel(logging.ERROR)

    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)

    if not patch_plan_dir.exists() or not patch_plan_dir.is_dir():
        rprint(f"{ERROR_BADGE} patch_plan_dir must be an existing directory")
        raise typer.Exit(1)

    async with dbdriver.start_session() as db:
        patch_runner = get_patch_runner(db=db)
        await patch_runner.apply(patch_plan_directory=patch_plan_dir)
        rprint(f"{SUCCESS_BADGE} Patch plan successfully applied")

    await dbdriver.close()


@patch_app.command(name="revert")
async def revert_patch_cmd(
    ctx: typer.Context,
    patch_plan_dir: Path = typer.Argument(help="Path to the directory containing a patch plan"),
    config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG"),
) -> None:
    """Revert a given patch plan"""
    logging.getLogger("infrahub").setLevel(logging.WARNING)
    logging.getLogger("neo4j").setLevel(logging.ERROR)
    logging.getLogger("prefect").setLevel(logging.ERROR)
    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    db = await context.init_db(retry=1)

    if not patch_plan_dir.exists() or not patch_plan_dir.is_dir():
        rprint(f"{ERROR_BADGE} patch_plan_dir must be an existing directory")
        raise typer.Exit(1)

    patch_runner = get_patch_runner(db=db)
    await patch_runner.revert(patch_plan_directory=patch_plan_dir)
    rprint(f"{SUCCESS_BADGE} Patch plan successfully reverted")

    await db.close()



from __future__ import annotations

import asyncio
import time
from typing import TYPE_CHECKING, Any

from ..graphql import Query
from .constants import FINAL_STATES
from .exceptions import TaskNotCompletedError, TaskNotFoundError, TooManyTasksError
from .models import Task, TaskFilter

if TYPE_CHECKING:
    from ..client import InfrahubClient, InfrahubClientSync


class InfraHubTaskManagerBase:
    @classmethod
    def _generate_query(
        cls,
        filters: TaskFilter | None = None,
        include_logs: bool = False,
        include_related_nodes: bool = False,
        offset: int | None = None,
        limit: int | None = None,
        count: bool = False,
    ) -> Query:
        query: dict[str, Any] = {
            "InfrahubTask": {
                "edges": {
                    "node": {
                        "id": None,
                        "title": None,
                        "state": None,
                        "progress": None,
                        "workflow": None,
                        "branch": None,
                        "created_at": None,
                        "updated_at": None,
                    }
                }
            }
        }

        if not filters and (offset or limit):
            filters = TaskFilter(offset=offset, limit=limit)
        elif filters and offset:
            filters.offset = offset
        elif filters and limit:
            filters.limit = limit

        if filters:
            query["InfrahubTask"]["@filters"] = filters.to_dict()

        if count:
            query["InfrahubTask"]["count"] = None

        if include_logs:
            query["InfrahubTask"]["edges"]["node"]["logs"] = {
                "edges": {
                    "node": {
                        "message": None,
                        "severity": None,
                        "timestamp": None,
                    }
                }
            }

        if include_related_nodes:
            query["InfrahubTask"]["edges"]["node"]["related_nodes"] = {"id": None, "kind": None}

        return Query(query=query)

    @classmethod
    def _generate_count_query(cls, filters: TaskFilter | None = None) -> Query:
        query: dict[str, Any] = {
            "InfrahubTask": {
                "count": None,
            }
        }
        if filters:
            query["InfrahubTask"]["@filters"] = filters.to_dict()

        return Query(query=query)


class InfrahubTaskManager(InfraHubTaskManagerBase):
    client: InfrahubClient

    def __init__(self, client: InfrahubClient) -> None:
        self.client = client

    async def count(self, filters: TaskFilter | None = None) -> int:
        """Count the number of tasks.

        Args:
            filters: The filter to apply to the tasks. Defaults to None.

        Returns:
            The number of tasks.
        """

        query = self._generate_count_query(filters=filters)
        response = await self.client.execute_graphql(
            query=query.render(convert_enum=False), tracker="query-tasks-count"
        )
        return int(response["InfrahubTask"]["count"])

    async def all(
        self,
        limit: int | None = None,
        offset: int | None = None,
        timeout: int | None = None,
        parallel: bool = False,
        include_logs: bool = False,
        include_related_nodes: bool = False,
    ) -> list[Task]:
        """Get all tasks.

        Args:
            limit: The maximum number of tasks to return. Defaults to None.
            offset: The offset to start the tasks from. Defaults to None.
            timeout: The timeout to wait for the tasks to complete. Defaults to None.
            parallel: Whether to query the tasks in parallel. Defaults to False.
            include_logs: Whether to include the logs in the tasks. Defaults to False.
            include_related_nodes: Whether to include the related nodes in the tasks. Defaults to False.

        Returns:
            A list of tasks.
        """

        return await self.filter(
            limit=limit,
            offset=offset,
            timeout=timeout,
            parallel=parallel,
            include_logs=include_logs,
            include_related_nodes=include_related_nodes,
        )

    async def filter(
        self,
        filter: TaskFilter | None = None,
        limit: int | None = None,
        offset: int | None = None,
        timeout: int | None = None,
        parallel: bool = False,
        include_logs: bool = False,
        include_related_nodes: bool = False,
    ) -> list[Task]:
        """Filter tasks.

        Args:
            filter: The filter to apply to the tasks. Defaults to None.
            limit: The maximum number of tasks to return. Defaults to None.
            offset: The offset to start the tasks from. Defaults to None.
            timeout: The timeout to wait for the tasks to complete. Defaults to None.
            parallel: Whether to query the tasks in parallel. Defaults to False.
            include_logs: Whether to include the logs in the tasks. Defaults to False.
            include_related_nodes: Whether to include the related nodes in the tasks. Defaults to False.

        Returns:
            A list of tasks.
        """
        if filter is None:
            filter = TaskFilter()

        if limit:
            tasks, _ = await self.process_page(
                self.client, self._generate_query(filters=filter, offset=offset, limit=limit, count=False), 1, timeout
            )
            return tasks

        if parallel:
            return await self.process_batch(
                filters=filter,
                timeout=timeout,
                include_logs=include_logs,
                include_related_nodes=include_related_nodes,
            )

        return await self.process_non_batch(
            filters=filter,
            offset=offset,
            limit=limit,
            timeout=timeout,
            include_logs=include_logs,
            include_related_nodes=include_related_nodes,
        )

    async def get(self, id: str, include_logs: bool = False, include_related_nodes: bool = False) -> Task:
        tasks = await self.filter(
            filter=TaskFilter(ids=[id]),
            include_logs=include_logs,
            include_related_nodes=include_related_nodes,
            parallel=False,
        )
        if not tasks:
            raise TaskNotFoundError(id=id)

        if len(tasks) != 1:
            raise TooManyTasksError(expected_id=id, received_ids=[task.id for task in tasks])

        return tasks[0]

    async def wait_for_completion(self, id: str, interval: int = 1, timeout: int = 60) -> Task:
        """Wait for a task to complete.

        Args:
            id: The id of the task to wait for.
            interval: The interval to check the task state. Defaults to 1.
            timeout: The timeout to wait for the task to complete. Defaults to 60.

        Raises:
            TaskNotCompletedError: The task did not complete in the given timeout.

        Returns:
            The task object.
        """
        for _ in range(timeout // interval):
            task = await self.get(id=id)
            if task.state in FINAL_STATES:
                return task
            await asyncio.sleep(interval)
        raise TaskNotCompletedError(id=id, message=f"Task {id} did not complete in {timeout} seconds")

    @staticmethod
    async def process_page(
        client: InfrahubClient, query: Query, page_number: int, timeout: int | None = None
    ) -> tuple[list[Task], int | None]:
        """Process a single page of results.

        Args:
            client: The client to use to execute the query.
            query: The query to execute.
            page_number: The page number to process.
            timeout: The timeout to wait for the query to complete. Defaults to None.

        Returns:
            A tuple containing a list of tasks and the count of tasks.
        """

        response = await client.execute_graphql(
            query=query.render(convert_enum=False),
            tracker=f"query-tasks-page{page_number}",
            timeout=timeout,
        )
        count = response["InfrahubTask"].get("count", None)
        return [Task.from_graphql(task["node"]) for task in response["InfrahubTask"]["edges"]], count

    async def process_batch(
        self,
        filters: TaskFilter | None = None,
        timeout: int | None = None,
        include_logs: bool = False,
        include_related_nodes: bool = False,
    ) -> list[Task]:
        """Process queries in parallel mode."""
        pagination_size = self.client.pagination_size
        tasks = []
        batch_process = await self.client.create_batch()
        count = await self.count(filters=filters)
        total_pages = (count + pagination_size - 1) // pagination_size

        for page_number in range(1, total_pages + 1):
            page_offset = (page_number - 1) * pagination_size
            query = self._generate_query(
                filters=filters,
                offset=page_offset,
                limit=pagination_size,
                include_logs=include_logs,
                include_related_nodes=include_related_nodes,
                count=False,
            )
            batch_process.add(
                task=self.process_page, client=self.client, query=query, page_number=page_number, timeout=timeout
            )

        async for _, (new_tasks, _) in batch_process.execute():
            tasks.extend(new_tasks)

        return tasks

    async def process_non_batch(
        self,
        filters: TaskFilter | None = None,
        offset: int | None = None,
        limit: int | None = None,
        timeout: int | None = None,
        include_logs: bool = False,
        include_related_nodes: bool = False,
    ) -> list[Task]:
        """Process queries without parallel mode."""
        tasks = []
        has_remaining_items = True
        page_number = 1

        while has_remaining_items:
            page_offset = (page_number - 1) * self.client.pagination_size
            query = self._generate_query(
                filters=filters,
                offset=page_offset,
                limit=self.client.pagination_size,
                include_logs=include_logs,
                include_related_nodes=include_related_nodes,
                count=True,
            )
            new_tasks, count = await self.process_page(
                client=self.client, query=query, page_number=page_number, timeout=timeout
            )
            if count is None:
                raise ValueError("Count is None, a value must be retrieve from the query")

            tasks.extend(new_tasks)
            remaining_items = count - (page_offset + self.client.pagination_size)
            if remaining_items < 0 or offset is not None or limit is not None:
                has_remaining_items = False
            page_number += 1
        return tasks


class InfrahubTaskManagerSync(InfraHubTaskManagerBase):
    client: InfrahubClientSync

    def __init__(self, client: InfrahubClientSync) -> None:
        self.client = client

    def count(self, filters: TaskFilter | None = None) -> int:
        """Count the number of tasks.

        Args:
            filters: The filter to apply to the tasks. Defaults to None.

        Returns:
            The number of tasks.
        """

        query = self._generate_count_query(filters=filters)
        response = self.client.execute_graphql(query=query.render(convert_enum=False), tracker="query-tasks-count")
        return int(response["InfrahubTask"]["count"])

    def all(
        self,
        limit: int | None = None,
        offset: int | None = None,
        timeout: int | None = None,
        parallel: bool = False,
        include_logs: bool = False,
        include_related_nodes: bool = False,
    ) -> list[Task]:
        """Get all tasks.

        Args:
            limit: The maximum number of tasks to return. Defaults to None.
            offset: The offset to start the tasks from. Defaults to None.
            timeout: The timeout to wait for the tasks to complete. Defaults to None.
            parallel: Whether to query the tasks in parallel. Defaults to False.
            include_logs: Whether to include the logs in the tasks. Defaults to False.
            include_related_nodes: Whether to include the related nodes in the tasks. Defaults to False.

        Returns:
            A list of tasks.
        """

        return self.filter(
            limit=limit,
            offset=offset,
            timeout=timeout,
            parallel=parallel,
            include_logs=include_logs,
            include_related_nodes=include_related_nodes,
        )

    def filter(
        self,
        filter: TaskFilter | None = None,
        limit: int | None = None,
        offset: int | None = None,
        timeout: int | None = None,
        parallel: bool = False,
        include_logs: bool = False,
        include_related_nodes: bool = False,
    ) -> list[Task]:
        """Filter tasks.

        Args:
            filter: The filter to apply to the tasks. Defaults to None.
            limit: The maximum number of tasks to return. Defaults to None.
            offset: The offset to start the tasks from. Defaults to None.
            timeout: The timeout to wait for the tasks to complete. Defaults to None.
            parallel: Whether to query the tasks in parallel. Defaults to False.
            include_logs: Whether to include the logs in the tasks. Defaults to False.
            include_related_nodes: Whether to include the related nodes in the tasks. Defaults to False.

        Returns:
            A list of tasks.
        """
        if filter is None:
            filter = TaskFilter()

        if limit:
            tasks, _ = self.process_page(
                self.client, self._generate_query(filters=filter, offset=offset, limit=limit, count=False), 1, timeout
            )
            return tasks

        if parallel:
            return self.process_batch(
                filters=filter,
                timeout=timeout,
                include_logs=include_logs,
                include_related_nodes=include_related_nodes,
            )

        return self.process_non_batch(
            filters=filter,
            offset=offset,
            limit=limit,
            timeout=timeout,
            include_logs=include_logs,
            include_related_nodes=include_related_nodes,
        )

    def get(self, id: str, include_logs: bool = False, include_related_nodes: bool = False) -> Task:
        tasks = self.filter(
            filter=TaskFilter(ids=[id]),
            include_logs=include_logs,
            include_related_nodes=include_related_nodes,
            parallel=False,
        )
        if not tasks:
            raise TaskNotFoundError(id=id)

        if len(tasks) != 1:
            raise TooManyTasksError(expected_id=id, received_ids=[task.id for task in tasks])

        return tasks[0]

    def wait_for_completion(self, id: str, interval: int = 1, timeout: int = 60) -> Task:
        """Wait for a task to complete.

        Args:
            id: The id of the task to wait for.
            interval: The interval to check the task state. Defaults to 1.
            timeout: The timeout to wait for the task to complete. Defaults to 60.

        Raises:
            TaskNotCompletedError: The task did not complete in the given timeout.

        Returns:
            The task object.
        """
        for _ in range(timeout // interval):
            task = self.get(id=id)
            if task.state in FINAL_STATES:
                return task
            time.sleep(interval)
        raise TaskNotCompletedError(id=id, message=f"Task {id} did not complete in {timeout} seconds")

    @staticmethod
    def process_page(
        client: InfrahubClientSync, query: Query, page_number: int, timeout: int | None = None
    ) -> tuple[list[Task], int | None]:
        """Process a single page of results.

        Args:
            client: The client to use to execute the query.
            query: The query to execute.
            page_number: The page number to process.
            timeout: The timeout to wait for the query to complete. Defaults to None.

        Returns:
            A tuple containing a list of tasks and the count of tasks.
        """

        response = client.execute_graphql(
            query=query.render(convert_enum=False),
            tracker=f"query-tasks-page{page_number}",
            timeout=timeout,
        )
        count = response["InfrahubTask"].get("count", None)
        return [Task.from_graphql(task["node"]) for task in response["InfrahubTask"]["edges"]], count

    def process_batch(
        self,
        filters: TaskFilter | None = None,
        timeout: int | None = None,
        include_logs: bool = False,
        include_related_nodes: bool = False,
    ) -> list[Task]:
        """Process queries in parallel mode."""
        pagination_size = self.client.pagination_size
        tasks = []
        batch_process = self.client.create_batch()
        count = self.count(filters=filters)
        total_pages = (count + pagination_size - 1) // pagination_size

        for page_number in range(1, total_pages + 1):
            page_offset = (page_number - 1) * pagination_size
            query = self._generate_query(
                filters=filters,
                offset=page_offset,
                limit=pagination_size,
                include_logs=include_logs,
                include_related_nodes=include_related_nodes,
                count=False,
            )
            batch_process.add(
                task=self.process_page, client=self.client, query=query, page_number=page_number, timeout=timeout
            )

        for _, (new_tasks, _) in batch_process.execute():
            tasks.extend(new_tasks)

        return tasks

    def process_non_batch(
        self,
        filters: TaskFilter | None = None,
        offset: int | None = None,
        limit: int | None = None,
        timeout: int | None = None,
        include_logs: bool = False,
        include_related_nodes: bool = False,
    ) -> list[Task]:
        """Process queries without parallel mode."""
        tasks = []
        has_remaining_items = True
        page_number = 1

        while has_remaining_items:
            page_offset = (page_number - 1) * self.client.pagination_size
            query = self._generate_query(
                filters=filters,
                offset=page_offset,
                limit=self.client.pagination_size,
                include_logs=include_logs,
                include_related_nodes=include_related_nodes,
                count=True,
            )
            new_tasks, count = self.process_page(
                client=self.client, query=query, page_number=page_number, timeout=timeout
            )
            if count is None:
                raise ValueError("Count is None, a value must be retrieve from the query")

            tasks.extend(new_tasks)
            remaining_items = count - (page_offset + self.client.pagination_size)
            if remaining_items < 0 or offset is not None or limit is not None:
                has_remaining_items = False
            page_number += 1
        return tasks



from infrahub.core.validators.attribute.min_max import AttributeNumberChecker

from .attribute.choices import AttributeChoicesChecker
from .attribute.enum import AttributeEnumChecker
from .attribute.kind import AttributeKindChecker
from .attribute.length import AttributeLengthChecker
from .attribute.number_pool import AttributeNumberPoolChecker
from .attribute.optional import AttributeOptionalChecker
from .attribute.regex import AttributeRegexChecker
from .attribute.unique import AttributeUniquenessChecker
from .enum import ConstraintIdentifier
from .interface import ConstraintCheckerInterface
from .node.attribute import NodeAttributeAddChecker
from .node.generate_profile import NodeGenerateProfileChecker
from .node.hierarchy import NodeHierarchyChecker
from .node.inherit_from import NodeInheritFromChecker
from .node.relationship import NodeRelationshipAddChecker
from .relationship.count import RelationshipCountChecker
from .relationship.optional import RelationshipOptionalChecker
from .relationship.peer import RelationshipPeerChecker, RelationshipPeerParentChecker
from .uniqueness.checker import UniquenessChecker

CONSTRAINT_VALIDATOR_MAP: dict[str, type[ConstraintCheckerInterface] | None] = {
    "attribute.kind.update": AttributeKindChecker,
    "attribute.regex.update": AttributeRegexChecker,
    ConstraintIdentifier.ATTRIBUTE_PARAMETERS_REGEX_UPDATE.value: AttributeRegexChecker,
    "attribute.enum.update": AttributeEnumChecker,
    "attribute.min_length.update": AttributeLengthChecker,
    "attribute.max_length.update": AttributeLengthChecker,
    ConstraintIdentifier.ATTRIBUTE_PARAMETERS_MIN_LENGTH_UPDATE.value: AttributeLengthChecker,
    ConstraintIdentifier.ATTRIBUTE_PARAMETERS_MAX_LENGTH_UPDATE.value: AttributeLengthChecker,
    ConstraintIdentifier.ATTRIBUTE_PARAMETERS_MIN_VALUE_UPDATE.value: AttributeNumberChecker,
    ConstraintIdentifier.ATTRIBUTE_PARAMETERS_MAX_VALUE_UPDATE.value: AttributeNumberChecker,
    ConstraintIdentifier.ATTRIBUTE_PARAMETERS_EXCLUDED_VALUES_UPDATE.value: AttributeNumberChecker,
    ConstraintIdentifier.ATTRIBUTE_PARAMETERS_START_RANGE_UPDATE: AttributeNumberPoolChecker,
    ConstraintIdentifier.ATTRIBUTE_PARAMETERS_END_RANGE_UPDATE: AttributeNumberPoolChecker,
    "attribute.unique.update": AttributeUniquenessChecker,
    "attribute.optional.update": AttributeOptionalChecker,
    "attribute.choices.update": AttributeChoicesChecker,
    "relationship.peer.update": RelationshipPeerChecker,
    "relationship.cardinality.update": RelationshipCountChecker,
    "relationship.optional.update": RelationshipOptionalChecker,
    "relationship.min_count.update": RelationshipCountChecker,
    "relationship.max_count.update": RelationshipCountChecker,
    "relationship.common_parent.update": RelationshipPeerParentChecker,
    "node.inherit_from.update": NodeInheritFromChecker,
    "node.uniqueness_constraints.update": UniquenessChecker,
    "node.parent.update": NodeHierarchyChecker,
    "node.children.update": NodeHierarchyChecker,
    "node.generate_profile.update": NodeGenerateProfileChecker,
    "node.attribute.add": NodeAttributeAddChecker,
    "node.relationship.add": NodeRelationshipAddChecker,
}



from infrahub.core.validators.relationship.peer import RelationshipPeerParentChecker
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class SchemaRelationshipPeerParentConstraintDependency(DependencyBuilder[RelationshipPeerParentChecker]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> RelationshipPeerParentChecker:
        return RelationshipPeerParentChecker(db=context.db, branch=context.branch)



from typing import Any, Generator

from infrahub.core.query import Query, QueryType
from infrahub.database import InfrahubDatabase


class EnrichedDiffFieldSpecifiersQuery(Query):
    name = "enriched_diff_field_specifiers"
    type = QueryType.READ

    def __init__(self, diff_id: str, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.diff_id = diff_id

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params["diff_id"] = self.diff_id
        query = """
CALL () {
    MATCH (root:DiffRoot {uuid: $diff_id})-[:DIFF_HAS_NODE]->(node:DiffNode)-[:DIFF_HAS_ATTRIBUTE]->(attr:DiffAttribute)
    WHERE (root.is_merged IS NULL OR root.is_merged <> TRUE)
    RETURN node.uuid AS node_uuid, node.kind AS node_kind, attr.name AS field_name
    UNION
    MATCH (root:DiffRoot {uuid: $diff_id})-[:DIFF_HAS_NODE]->(node:DiffNode)-[:DIFF_HAS_RELATIONSHIP]->(rel:DiffRelationship)
    WHERE (root.is_merged IS NULL OR root.is_merged <> TRUE)
    RETURN node.uuid AS node_uuid, node.kind AS node_kind, rel.identifier AS field_name
}
        """
        self.add_to_query(query=query)
        self.return_labels = ["node_uuid", "node_kind", "field_name"]
        self.order_by = ["node_uuid", "node_kind", "field_name"]

    def get_node_field_specifier_tuples(self) -> Generator[tuple[str, str, str], None, None]:
        for result in self.get_results():
            node_uuid = result.get_as_str("node_uuid")
            node_kind = result.get_as_str("node_kind")
            field_name = result.get_as_str("field_name")
            if node_uuid and node_kind and field_name:
                yield (node_uuid, node_kind, field_name)



from collections import defaultdict

from infrahub.core.branch.models import Branch
from infrahub.core.constants import DiffAction, InfrahubKind
from infrahub.database import InfrahubDatabase

from ..model.diff import ArtifactTarget, BranchDiffArtifact, BranchDiffArtifactStorage
from ..payload_builder import get_display_labels_per_kind
from ..query.artifact import ArtifactDiffQuery


class ArtifactDiffCalculator:
    def __init__(self, db: InfrahubDatabase) -> None:
        self.db = db

    async def calculate(self, source_branch: Branch, target_branch: Branch) -> list[BranchDiffArtifact]:
        artifact_schema = self.db.schema.get(name=InfrahubKind.ARTIFACT, branch=source_branch, duplicate=False)
        target_rel = artifact_schema.get_relationship(name="object")
        definition_rel = artifact_schema.get_relationship(name="definition")
        query = await ArtifactDiffQuery.init(
            db=self.db,
            branch=source_branch,
            target_branch=target_branch,
            target_rel_identifier=target_rel.get_identifier(),
            definition_rel_identifier=definition_rel.get_identifier(),
        )
        await query.execute(db=self.db)

        artifact_diffs = []
        for result in query.get_results():
            source_artifact_node = result.get_node(label="source_artifact")
            artifact_id = str(source_artifact_node.get("uuid"))
            try:
                target_node = result.get_node(label="target_node")
                artifact_target = ArtifactTarget(
                    id=str(target_node.get("uuid")),
                    kind=str(target_node.get("kind")),
                    display_label=None,
                )
            except ValueError:
                artifact_target = None
            source_storage_id = result.get_as_str("source_storage_id")
            source_checksum = result.get_as_str("source_checksum")
            new_storage: BranchDiffArtifactStorage | None = None
            if source_storage_id and source_checksum:
                new_storage = BranchDiffArtifactStorage(
                    storage_id=source_storage_id,
                    checksum=source_checksum,
                )
            target_storage_id = result.get_as_str("target_storage_id")
            target_checksum = result.get_as_str("target_checksum")
            old_storage: BranchDiffArtifactStorage | None = None
            if target_storage_id and target_checksum:
                old_storage = BranchDiffArtifactStorage(
                    storage_id=target_storage_id,
                    checksum=target_checksum,
                )
            artifact_diffs.append(
                BranchDiffArtifact(
                    branch=source_branch.name,
                    id=artifact_id,
                    display_label=None,
                    action=DiffAction.UPDATED if old_storage else DiffAction.ADDED,
                    target=artifact_target,
                    item_new=new_storage,
                    item_previous=old_storage,
                )
            )
        await self._add_display_labels(source_branch=source_branch, artifact_diffs=artifact_diffs)

        return artifact_diffs

    async def _add_display_labels(self, source_branch: Branch, artifact_diffs: list[BranchDiffArtifact]) -> None:
        ids_by_kind: dict[str, set[str]] = defaultdict(set)
        for artifact_diff in artifact_diffs:
            ids_by_kind[InfrahubKind.ARTIFACT].add(artifact_diff.id)
            if artifact_diff.target:
                ids_by_kind[artifact_diff.target.kind].add(artifact_diff.target.id)

        display_labels_by_id: dict[str, str] = {}
        for kind, node_ids in ids_by_kind.items():
            display_label_map = await get_display_labels_per_kind(
                db=self.db, branch_name=source_branch.name, kind=kind, ids=list(node_ids)
            )
            display_labels_by_id.update(display_label_map)

        for artifact_diff in artifact_diffs:
            artifact_diff.display_label = display_labels_by_id.get(artifact_diff.id)
            if not artifact_diff.target:
                continue
            artifact_diff.target.display_label = display_labels_by_id.get(artifact_diff.target.id)
            if artifact_diff.target.display_label and artifact_diff.display_label:
                artifact_diff.display_label = f"{artifact_diff.target.display_label} - {artifact_diff.display_label}"



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub import lock
from infrahub.core import registry
from infrahub.core.query.resource_manager import NumberPoolGetReserved, NumberPoolGetUsed, NumberPoolSetReserved
from infrahub.core.schema.attribute_parameters import NumberAttributeParameters
from infrahub.exceptions import PoolExhaustedError

from .. import Node
from ..lock_utils import RESOURCE_POOL_LOCK_NAMESPACE

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.schema import AttributeSchema
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase


class CoreNumberPool(Node):
    def get_attribute_nb_excluded_values(self) -> int:
        """Returns the number of excluded values for the attribute of the number pool."""

        pool_node = registry.schema.get(name=self.node.value)  # type: ignore [attr-defined]
        attribute = [attribute for attribute in pool_node.attributes if attribute.name == self.node_attribute.value][0]  # type: ignore [attr-defined]
        if not isinstance(attribute.parameters, NumberAttributeParameters):
            return 0

        sum_excluded_values = 0
        excluded_ranges = attribute.parameters.get_excluded_ranges()
        for start_range, end_range in excluded_ranges:
            sum_excluded_values += end_range - start_range + 1

        res = len(attribute.parameters.get_excluded_single_values()) + sum_excluded_values
        return res

    async def get_used(
        self,
        db: InfrahubDatabase,
        branch: Branch,
    ) -> list[int]:
        """Returns a list of used numbers in the pool."""

        query = await NumberPoolGetUsed.init(db=db, branch=branch, pool=self, branch_agnostic=True)
        await query.execute(db=db)
        used = [result.value for result in query.iter_results()]
        return [item for item in used if item is not None]

    async def reserve(self, db: InfrahubDatabase, number: int, identifier: str, at: Timestamp | None = None) -> None:
        """Reserve a number in the pool for a specific identifier."""

        query = await NumberPoolSetReserved.init(
            db=db, pool_id=self.get_id(), identifier=identifier, reserved=number, at=at
        )
        await query.execute(db=db)

    async def get_resource(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        node: Node,
        attribute: AttributeSchema,
        identifier: str | None = None,
        at: Timestamp | None = None,
    ) -> int:
        async with lock.registry.get(name=self.get_id(), namespace=RESOURCE_POOL_LOCK_NAMESPACE):
            # NOTE: ideally we should use the HFID as the identifier (if available)
            # one of the challenge with using the HFID is that it might change over time
            # so we need to ensure that the identifier is stable, or we need to handle the case where the identifier changes
            identifier = identifier or node.get_id()

            # Check if there is already a resource allocated with this identifier
            # if not, pull all existing number and allocate the next available
            # TODO add support for branch, if the node is reserved with this id in another branch we should return an error
            query_get = await NumberPoolGetReserved.init(db=db, branch=branch, pool_id=self.id, identifier=identifier)
            await query_get.execute(db=db)
            reservation = query_get.get_reservation()
            if reservation is not None:
                return reservation

            # If we have not returned a value we need to find one if avaiable
            number = await self.get_next(db=db, branch=branch, attribute=attribute)
            await self.reserve(db=db, number=number, identifier=identifier, at=at)
            return number

    async def get_next(self, db: InfrahubDatabase, branch: Branch, attribute: AttributeSchema) -> int:
        taken = await self.get_used(db=db, branch=branch)

        next_number = find_next_free(
            start=self.start_range.value,  # type: ignore[attr-defined]
            end=self.end_range.value,  # type: ignore[attr-defined]
            taken=taken,
            parameters=attribute.parameters if isinstance(attribute.parameters, NumberAttributeParameters) else None,
        )
        if next_number is None:
            raise PoolExhaustedError("There are no more values available in this pool.")

        return next_number

    async def get_next_many(
        self, db: InfrahubDatabase, quantity: int, branch: Branch, attribute: AttributeSchema
    ) -> list[int]:
        taken = await self.get_used(db=db, branch=branch)

        allocated: list[int] = []

        for _ in range(quantity):
            next_number = find_next_free(
                start=self.start_range.value,  # type: ignore[attr-defined]
                end=self.end_range.value,  # type: ignore[attr-defined]
                taken=list(set(taken) | set(allocated)),
                parameters=attribute.parameters
                if isinstance(attribute.parameters, NumberAttributeParameters)
                else None,
            )
            if next_number is None:
                raise PoolExhaustedError(
                    f"There are no more values available in this pool, couldn't allocate {quantity} values, only {len(allocated)} available."
                )

            allocated.append(next_number)

        return allocated


def find_next_free(start: int, end: int, taken: list[int], parameters: NumberAttributeParameters | None) -> int | None:
    used_set = set(taken)

    for num in range(start, end + 1):
        if num not in used_set:
            if parameters is None or parameters.is_valid_value(num):
                return num

    return None



import ipaddress
from ipaddress import IPv4Interface
from typing import TYPE_CHECKING, Any

from graphene import InputObjectType, Mutation
from graphql import GraphQLResolveInfo
from typing_extensions import Self

from infrahub import lock
from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.constants import InfrahubKind
from infrahub.core.ipam.reconciler import IpamReconciler
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.schema import NodeSchema
from infrahub.database import InfrahubDatabase, retry_db_transaction
from infrahub.exceptions import NodeNotFoundError, ValidationError
from infrahub.lock import InfrahubMultiLock
from infrahub.log import get_logger

from ...core.node.create import create_node
from ...core.node.lock_utils import build_object_lock_name, get_lock_names_on_object_mutation
from .main import DeleteResult, InfrahubMutationMixin, InfrahubMutationOptions, build_graphql_response
from .node_getter.by_default_filter import MutationNodeGetterByDefaultFilter

if TYPE_CHECKING:
    from infrahub.graphql.initialization import GraphqlContext

log = get_logger()


async def validate_namespace(
    db: InfrahubDatabase,
    branch: Branch | str | None,
    data: InputObjectType,
    existing_namespace_id: str | None = None,
) -> str:
    """Validate or set (if not present) the namespace to pass to the mutation and return its ID."""
    namespace_id: str | None = None
    if "ip_namespace" not in data or not data["ip_namespace"]:
        namespace_id = existing_namespace_id or registry.default_ipnamespace
        data["ip_namespace"] = {"id": namespace_id}
    elif "id" in data["ip_namespace"]:
        namespace = await registry.manager.get_one(
            db=db, branch=branch, kind=InfrahubKind.IPNAMESPACE, id=data["ip_namespace"]["id"]
        )
        namespace_id = namespace.id
    elif "hfid" in data["ip_namespace"]:
        namespace = await registry.manager.get_one_by_hfid(
            db=db, branch=branch, kind=InfrahubKind.IPNAMESPACE, hfid=data["ip_namespace"]["hfid"]
        )
        namespace_id = namespace.id
    else:
        raise ValidationError(
            "A valid ip_namespace must be provided or ip_namespace should be left empty in order to use the default value."
        )
    return namespace_id


class InfrahubIPNamespaceMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(
        cls,
        schema: NodeSchema,
        _meta: Any | None = None,
        **options: dict[str, Any],
    ) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)
        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    async def mutate_delete(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
    ) -> DeleteResult:
        if data["id"] == registry.default_ipnamespace:
            raise ValueError("Cannot delete default IPAM namespace")

        return await super().mutate_delete(info=info, data=data, branch=branch)


class InfrahubIPAddressMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(
        cls,
        schema: NodeSchema,
        _meta: Any | None = None,
        **options: dict[str, Any],
    ) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)
        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @staticmethod
    def _get_lock_names(namespace_id: str, branch: Branch) -> list[str]:
        if not branch.is_default:
            # Do not lock on other branches as reconciliation will be performed at least when merging in main branch.
            return []
        return [build_object_lock_name(InfrahubKind.IPADDRESS + "_" + namespace_id)]

    @classmethod
    async def _mutate_create_object_and_reconcile(
        cls,
        data: InputObjectType,
        branch: Branch,
        db: InfrahubDatabase,
        ip_address: IPv4Interface | ipaddress.IPv6Interface,
        namespace_id: str,
    ) -> Node:
        address = await create_node(
            data=dict(data),
            db=db,
            branch=branch,
            schema=cls._meta.active_schema,
        )

        reconciler = IpamReconciler(db=db, branch=branch)
        reconciled_address = await reconciler.reconcile(
            ip_value=ip_address, namespace=namespace_id, node_uuid=address.get_id()
        )
        return reconciled_address

    @classmethod
    @retry_db_transaction(name="ipaddress_create")
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context
        db = database or graphql_context.db
        ip_address = ipaddress.ip_interface(data["address"]["value"])
        namespace_id = await validate_namespace(db=db, branch=branch, data=data)

        lock_names = cls._get_lock_names(namespace_id, branch)
        async with InfrahubMultiLock(lock_registry=lock.registry, locks=lock_names):
            async with db.start_transaction() as dbt:
                reconciled_address = await cls._mutate_create_object_and_reconcile(
                    data=data, branch=branch, db=dbt, ip_address=ip_address, namespace_id=namespace_id
                )
                graphql_response = await build_graphql_response(info=info, db=dbt, obj=reconciled_address)

        return reconciled_address, cls(**graphql_response)

    @classmethod
    async def _mutate_update_object_and_reconcile(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        db: InfrahubDatabase,
        address: Node,
        namespace_id: str,
    ) -> Node:
        address = await cls.mutate_update_object(db=db, info=info, data=data, branch=branch, obj=address)
        reconciler = IpamReconciler(db=db, branch=branch)
        ip_address = ipaddress.ip_interface(address.address.value)
        reconciled_address = await reconciler.reconcile(
            ip_value=ip_address, node_uuid=address.get_id(), namespace=namespace_id
        )
        return reconciled_address

    @classmethod
    @retry_db_transaction(name="ipaddress_update")
    async def mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,
        node: Node | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context
        db = database or graphql_context.db

        address = node or await NodeManager.get_one_by_id_or_default_filter(
            db=db,
            kind=cls._meta.schema.kind,
            id=data.get("id"),
            branch=branch,
            include_owner=True,
            include_source=True,
        )
        namespace = await address.ip_namespace.get_peer(db)
        namespace_id = await validate_namespace(db=db, branch=branch, data=data, existing_namespace_id=namespace.id)

        # Prepare a clone to compute locks without triggering pool allocations
        preview_obj = await NodeManager.get_one_by_id_or_default_filter(
            db=db,
            kind=address.get_kind(),
            id=address.get_id(),
            branch=branch,
        )
        await preview_obj.from_graphql(db=db, data=data, process_pools=False)

        schema_branch = db.schema.get_schema_branch(name=branch.name)
        lock_names = get_lock_names_on_object_mutation(node=preview_obj, schema_branch=schema_branch)

        namespace_lock_names = cls._get_lock_names(namespace_id, branch)
        async with InfrahubMultiLock(lock_registry=lock.registry, locks=namespace_lock_names):
            # FIXME: do not lock when data does not contain uniqueness constraint fields or resource pool allocations
            async with InfrahubMultiLock(lock_registry=lock.registry, locks=lock_names, metrics=False):
                async with db.start_transaction() as dbt:
                    reconciled_address = await cls._mutate_update_object_and_reconcile(
                        info=info, data=data, branch=branch, address=address, namespace_id=namespace_id, db=dbt
                    )

                    result = await cls.mutate_update_to_graphql(db=dbt, info=info, obj=reconciled_address)

        return address, result

    @classmethod
    async def mutate_upsert(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        node_getter_default_filter: MutationNodeGetterByDefaultFilter,
        database: InfrahubDatabase | None = None,
    ) -> tuple[Node, Self, bool]:
        graphql_context: GraphqlContext = info.context
        db = database or graphql_context.db

        await validate_namespace(db=db, branch=branch, data=data)
        prefix, result, created = await super().mutate_upsert(
            info=info, data=data, branch=branch, node_getter_default_filter=node_getter_default_filter, database=db
        )

        return prefix, result, created

    @classmethod
    async def mutate_delete(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
    ) -> DeleteResult:
        return await super().mutate_delete(info=info, data=data, branch=branch)


class InfrahubIPPrefixMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(
        cls,
        schema: NodeSchema,
        _meta: Any | None = None,
        **options: dict[str, Any],
    ) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)
        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @staticmethod
    def _get_lock_names(namespace_id: str) -> list[str]:
        # IPPrefix has some cardinality-one relationships involved (parent/child/ip_address),
        # so we need to lock on any branch to avoid creating multiple peers for these relationships
        # during concurrent ipam reconciliations.
        return [build_object_lock_name(InfrahubKind.IPPREFIX + "_" + namespace_id)]

    @classmethod
    async def _mutate_create_object_and_reconcile(
        cls,
        data: InputObjectType,
        branch: Branch,
        db: InfrahubDatabase,
        namespace_id: str,
    ) -> Node:
        prefix = await create_node(
            data=dict(data),
            db=db,
            branch=branch,
            schema=cls._meta.active_schema,
        )
        return await cls._reconcile_prefix(
            branch=branch, db=db, prefix=prefix, namespace_id=namespace_id, is_delete=False
        )

    @classmethod
    @retry_db_transaction(name="ipprefix_create")
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context
        db = database or graphql_context.db
        namespace_id = await validate_namespace(db=db, branch=branch, data=data)

        lock_names = cls._get_lock_names(namespace_id)
        async with InfrahubMultiLock(lock_registry=lock.registry, locks=lock_names):
            async with db.start_transaction() as dbt:
                reconciled_prefix = await cls._mutate_create_object_and_reconcile(
                    data=data, branch=branch, db=dbt, namespace_id=namespace_id
                )

            graphql_response = await build_graphql_response(info=info, db=dbt, obj=reconciled_prefix)

        return reconciled_prefix, cls(**graphql_response)

    @classmethod
    async def _mutate_update_object_and_reconcile(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        db: InfrahubDatabase,
        prefix: Node,
        namespace_id: str,
    ) -> Node:
        prefix = await cls.mutate_update_object(db=db, info=info, data=data, branch=branch, obj=prefix)
        return await cls._reconcile_prefix(
            branch=branch, db=db, prefix=prefix, namespace_id=namespace_id, is_delete=False
        )

    @classmethod
    @retry_db_transaction(name="ipprefix_update")
    async def mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,
        node: Node | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context
        db = database or graphql_context.db

        prefix = node or await NodeManager.get_one_by_id_or_default_filter(
            db=db,
            kind=cls._meta.schema.kind,
            id=data.get("id"),
            branch=branch,
            include_owner=True,
            include_source=True,
        )
        namespace = await prefix.ip_namespace.get_peer(db)
        namespace_id = await validate_namespace(db=db, branch=branch, data=data, existing_namespace_id=namespace.id)

        # Prepare a clone to compute locks without triggering pool allocations
        preview_obj = await NodeManager.get_one_by_id_or_default_filter(
            db=db,
            kind=prefix.get_kind(),
            id=prefix.get_id(),
            branch=branch,
        )
        await preview_obj.from_graphql(db=db, data=data, process_pools=False)

        schema_branch = db.schema.get_schema_branch(name=branch.name)
        lock_names = get_lock_names_on_object_mutation(node=preview_obj, schema_branch=schema_branch)

        namespace_lock_names = cls._get_lock_names(namespace_id)
        async with InfrahubMultiLock(lock_registry=lock.registry, locks=namespace_lock_names):
            async with InfrahubMultiLock(lock_registry=lock.registry, locks=lock_names, metrics=False):
                async with db.start_transaction() as dbt:
                    reconciled_prefix = await cls._mutate_update_object_and_reconcile(
                        info=info, data=data, prefix=prefix, db=dbt, namespace_id=namespace_id, branch=branch
                    )
                    result = await cls.mutate_update_to_graphql(db=dbt, info=info, obj=reconciled_prefix)

        return prefix, result

    @classmethod
    async def mutate_upsert(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        node_getter_default_filter: MutationNodeGetterByDefaultFilter,
        database: InfrahubDatabase | None = None,
    ) -> tuple[Node, Self, bool]:
        graphql_context: GraphqlContext = info.context
        db = database or graphql_context.db

        await validate_namespace(db=db, branch=branch, data=data)
        prefix, result, created = await super().mutate_upsert(
            info=info, data=data, branch=branch, node_getter_default_filter=node_getter_default_filter, database=db
        )

        return prefix, result, created

    @classmethod
    async def _reconcile_prefix(
        cls,
        branch: Branch,
        db: InfrahubDatabase,
        prefix: Node,
        namespace_id: str,
        is_delete: bool,
    ) -> Node:
        reconciler = IpamReconciler(db=db, branch=branch)
        ip_network = ipaddress.ip_network(prefix.prefix.value)
        reconciled_prefix = await reconciler.reconcile(
            ip_value=ip_network, node_uuid=prefix.get_id(), namespace=namespace_id, is_delete=is_delete
        )
        return reconciled_prefix

    @classmethod
    @retry_db_transaction(name="ipprefix_delete")
    async def mutate_delete(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
    ) -> DeleteResult:
        graphql_context: GraphqlContext = info.context
        db = graphql_context.db

        prefix = await NodeManager.get_one(
            data.get("id"), graphql_context.db, branch=branch, prefetch_relationships=True
        )
        if not prefix:
            raise NodeNotFoundError(branch, cls._meta.schema.kind, data.get("id"))

        namespace_rels = await prefix.ip_namespace.get_relationships(db=db)
        namespace_id = namespace_rels[0].peer_id

        lock_names = cls._get_lock_names(namespace_id)
        async with InfrahubMultiLock(lock_registry=lock.registry, locks=lock_names):
            async with graphql_context.db.start_transaction() as dbt:
                reconciled_prefix = await cls._reconcile_prefix(
                    branch=branch, db=dbt, prefix=prefix, namespace_id=namespace_id, is_delete=True
                )
        ok = True

        return DeleteResult(node=reconciled_prefix, mutation=cls(ok=ok))



from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from infrahub_sdk.template import Jinja2Template

from infrahub.core.query.node import AttributeFromDB
from infrahub.core.schema import NodeSchema, ProfileSchema, TemplateSchema

from ..attribute import BaseAttribute, ListAttributeOptional, StringOptional

if TYPE_CHECKING:
    from infrahub.core.node import Node
    from infrahub.core.schema import NodeSchema, ProfileSchema, TemplateSchema
    from infrahub.core.schema.attribute_schema import AttributeSchema
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase

T = TypeVar("T")


class NodePropertyAttribute(Generic[T]):
    """A node property attribute is a construct that seats between a property and an attribute.

    View it as a property, set at the node level but stored in the database as an attribute. It usually is something computed from other components of
    a node, such as its attributes and its relationships.
    """

    def __init__(
        self,
        node_schema: NodeSchema | ProfileSchema | TemplateSchema,
        template: T | None,
        value: AttributeFromDB | T | None = None,
    ) -> None:
        self.node_schema = node_schema

        self.node_attributes: list[str] = []
        self.node_relationships: list[str] = []

        self.template = template
        self._value = value
        self._manually_assigned = False

        self.schema: AttributeSchema

        self.analyze_variables()

    def needs_update(self, fields: list[str] | None) -> bool:
        """Tell if this node property attribute must be recomputed given a list of updated fields of a node."""
        if self._manually_assigned or not fields:
            return True
        for field in fields:
            if field in self.node_attributes or field in self.node_relationships:
                return True

        return False

    @property
    def attribute_value(self) -> AttributeFromDB | dict[str, T | None]:
        if isinstance(self._value, AttributeFromDB):
            return self._value
        return {"value": self._value}

    def set_value(self, value: T | None, manually_assigned: bool = False) -> None:
        """Force the value of the node property attribute to the given one."""
        if isinstance(self._value, AttributeFromDB):
            self._value.value = value
        else:
            self._value = value

        if manually_assigned:
            self._manually_assigned = True

    def get_value(self, node: Node, at: Timestamp) -> T | None:
        if isinstance(self._value, AttributeFromDB):
            attr = self.get_node_attribute(node=node, at=at)
            return attr.value  # type: ignore

        return self._value

    @abstractmethod
    def analyze_variables(self) -> None: ...

    @abstractmethod
    async def compute(self, db: InfrahubDatabase, node: Node) -> None: ...

    @abstractmethod
    def get_node_attribute(self, node: Node, at: Timestamp) -> BaseAttribute: ...


class DisplayLabel(NodePropertyAttribute[str]):
    def __init__(
        self,
        node_schema: NodeSchema | ProfileSchema | TemplateSchema,
        template: str | None,
        value: AttributeFromDB | str | None = None,
    ) -> None:
        super().__init__(node_schema=node_schema, template=template, value=value)

        self.schema = node_schema.get_attribute(name="display_label")

    @property
    def is_jinja2_template(self) -> bool:
        if self.template is None:
            return False

        return any(c in self.template for c in "{}")

    def _analyze_plain_value(self) -> None:
        if self.template is None or "__" not in self.template:
            return

        items = self.template.split("__", maxsplit=1)
        if items[0] not in self.node_schema.attribute_names:
            raise ValueError(f"{items[0]} is not an attribute of {self.node_schema.kind}")

        self.node_attributes.append(items[0])

    def _analyze_jinja2_value(self) -> None:
        if self.template is None or not self.is_jinja2_template:
            return

        tpl = Jinja2Template(template=self.template)
        for variable in tpl.get_variables():
            items = variable.split("__", maxsplit=1)
            if items[0] in self.node_schema.attribute_names:
                self.node_attributes.append(items[0])
            elif items[0] in self.node_schema.relationship_names:
                self.node_relationships.append(items[0])
            else:
                raise ValueError(f"{items[0]} is neither an attribute or a relationship of {self.node_schema.kind}")

    def analyze_variables(self) -> None:
        """Look at variables used in the display label and record attributes and relationships required to compute it."""
        if not self.is_jinja2_template:
            self._analyze_plain_value()
        else:
            self._analyze_jinja2_value()

    async def compute(self, db: InfrahubDatabase, node: Node) -> None:
        """Update the display label value by recomputing it from the template."""
        if self.template is None or self._manually_assigned:
            return

        if node.get_schema() != self.node_schema:
            raise ValueError(
                f"display_label for schema {self.node_schema.kind} cannot be rendered for node {node.get_schema().kind} {node.id}"
            )

        if not self.is_jinja2_template:
            path_value = await node.get_path_value(db=db, path=self.template)
            # Use .value for enum to keep compat with old style display label
            self.set_value(value=str(path_value if not isinstance(path_value, Enum) else path_value.value))
            return

        jinja2_template = Jinja2Template(template=self.template)

        variables: dict[str, Any] = {}
        for variable in jinja2_template.get_variables():
            variables[variable] = await node.get_path_value(db=db, path=variable)

        self.set_value(value=await jinja2_template.render(variables=variables))

    def get_node_attribute(self, node: Node, at: Timestamp) -> StringOptional:
        """Return a node attribute that can be stored in the database for this display label and node."""
        return StringOptional(
            name="display_label",
            schema=self.schema,
            branch=node.get_branch(),
            at=at,
            node=node,
            data=self.attribute_value,
        )


class HumanFriendlyIdentifier(NodePropertyAttribute[list[str]]):
    def __init__(
        self,
        node_schema: NodeSchema | ProfileSchema | TemplateSchema,
        template: list[str] | None,
        value: AttributeFromDB | list[str] | None = None,
    ) -> None:
        super().__init__(node_schema=node_schema, template=template, value=value)

        self.schema = node_schema.get_attribute(name="human_friendly_id")

    def _analyze_single_variable(self, value: str) -> None:
        items = value.split("__", maxsplit=1)
        if items[0] in self.node_schema.attribute_names:
            self.node_attributes.append(items[0])
        elif items[0] in self.node_schema.relationship_names:
            self.node_relationships.append(items[0])
        else:
            raise ValueError(f"{items[0]} is neither an attribute or a relationship of {self.node_schema.kind}")

    def analyze_variables(self) -> None:
        """Look at variables used in the HFID and record attributes and relationships required to compute it."""
        for item in self.template or []:
            self._analyze_single_variable(value=item)

    async def compute(self, db: InfrahubDatabase, node: Node) -> None:
        """Update the HFID value by recomputing it from the template."""
        if self.template is None or self._manually_assigned:
            return

        if node.get_schema() != self.node_schema:
            raise ValueError(
                f"human_friendly_id for schema {self.node_schema.kind} cannot be computed for node {node.get_schema().kind} {node.id}"
            )

        value: list[str] = []
        for path in self.template:
            path_value = await node.get_path_value(db=db, path=path)
            # Use .value for enum to be consistent with display label
            value.append(path_value if not isinstance(path_value, Enum) else path_value.value)

        self.set_value(value=value)

    def get_node_attribute(self, node: Node, at: Timestamp) -> ListAttributeOptional:
        """Return a node attribute that can be stored in the database for this HFID and node."""
        return ListAttributeOptional(
            name="human_friendly_id",
            schema=self.schema,
            branch=node.get_branch(),
            at=at,
            node=node,
            data=self.attribute_value,
        )
